from requests import Session

from darwin.future.core.types.common import JSONType


def get_team_raw(session: Session, url: str) -> JSONType:
    """Gets the raw JSON response from a team endpoint

    Parameters:
        session (Session): Requests session to use
        url (str): URL to get

    Returns:
        JSONType: JSON response from the endpoint
    """
    response = session.get(url)
    response.raise_for_status()
    return response.json()



from __future__ import annotations

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONType
from darwin.future.data_objects.typing import UnknownType


def tag_items(
    client: ClientCore,
    team_slug: str,
    dataset_ids: int | list[int],
    tag_id: int,
    filters: dict[str, UnknownType],
) -> JSONType:
    """
    Adds tag annotation to all items slots matched by filters.

    Args:
        client (ClientCore): The Darwin Core client.
        team_slug (str): The team slug.
        dataset_ids (int | list[int]): The dataset ids.
        tag_id (int): The tag id.
        filters Dict[str, UnknownType]: The parameters of the filter.

    Returns:
        JSONType: The response data.
    """
    assert (
        filters
    ), "No parameters provided, please provide at least one non-dataset id filter"
    payload = {
        "filters": {
            "dataset_ids": (
                dataset_ids if isinstance(dataset_ids, list) else [dataset_ids]
            ),
            **filters,
        },
        "annotation_class_id": tag_id,
    }

    return client.post(f"/v2/teams/{team_slug}/items/slots/tags", data=payload)



from __future__ import annotations

import json
import os
from enum import Enum
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

from pydantic import field_validator

from darwin.future.data_objects.pydantic_base import DefaultDarwin

PropertyType = Literal[
    "multi_select",
    "single_select",
    "text",
    "attributes",
    "instance_id",
    "directional_vector",
]


class PropertyGranularity(str, Enum):
    section = "section"
    annotation = "annotation"
    item = "item"


class PropertyValue(DefaultDarwin):
    """
    Describes a single option for a property

    Attributes:
        value (str): Value of the option
        color (Optional[str]): Color of the option
        type (Optional[str]): Type of the option

    Validators:
        color (validator): Validates that the color is in rgba format
    """

    id: Optional[str] = None
    type: Literal["string"] = "string"
    value: Optional[str] = None
    color: str = "auto"

    @field_validator("color")
    @classmethod
    def validate_rgba(cls, v: str) -> str:
        if not v.startswith("rgba") and v != "auto":
            raise ValueError("Color must be in rgba format or 'auto'")
        return v

    def to_update_endpoint(self) -> Tuple[str, dict]:
        if self.id is None:
            raise ValueError("id must be set")
        updated_base = self.model_dump(exclude={"id", "type"})
        return self.id, updated_base


class FullProperty(DefaultDarwin):
    """
    Describes the property and all of the potential options that are associated with it

    Attributes:
        name (str): Name of the property
        type (str): Type of the property
        required (bool): If the property is required
        options (List[PropertyOption]): List of all options for the property
        granularity (PropertyGranularity): Granularity of the property

    """

    id: Optional[str] = None
    position: Optional[int] = None
    name: str
    type: PropertyType
    description: Optional[str] = None
    required: bool
    slug: Optional[str] = None
    team_id: Optional[int] = None
    annotation_class_id: Optional[int] = None
    property_values: Optional[List[PropertyValue]] = None
    granularity: PropertyGranularity
    dataset_ids: Optional[List[int]] = None
    options: Optional[List[PropertyValue]] = None
    granularity: PropertyGranularity = PropertyGranularity("section")

    def to_create_endpoint(
        self,
    ) -> dict:
        include_fields = {
            "name": True,
            "type": True,
            "required": True,
            "description": True,
            "granularity": True,
        }
        if self.type != "text":
            include_fields["property_values"] = {"__all__": {"value", "color", "type"}}
        if self.granularity != PropertyGranularity.item:
            if self.annotation_class_id is None:
                raise ValueError("annotation_class_id must be set")
            include_fields["annotation_class_id"] = True
        if self.dataset_ids is not None:
            include_fields["dataset_ids"] = True
        return self.model_dump(mode="json", include=include_fields)

    def to_update_endpoint(self) -> Tuple[str, dict]:
        if self.id is None:
            raise ValueError("id must be set")

        updated_base = self.to_create_endpoint()
        updated_base.pop("annotation_class_id", None)  # Can't update this field
        del updated_base["granularity"]  # Can't update this field
        return self.id, updated_base


class MetaDataClass(DefaultDarwin):
    """
    Metadata.json -> property mapping. Contains all properties for a class contained
    in the metadata.json file. Along with all options for each property that is associated
    with the class.

    Attributes:
        name (str): Name of the class
        type (str): Type of the class
        description (Optional[str]): Description of the class
        color (Optional[str]): Color of the class in the UI
        sub_types (Optional[List[str]]): Sub types of the class
        granularity:(PropertyGranularity): Granularity of the property
        properties (List[FullProperty]): List of all properties for the class with all options
    """

    name: str
    type: str
    description: Optional[str] = None
    color: Optional[str] = None
    sub_types: Optional[List[str]] = None
    granularity: PropertyGranularity = PropertyGranularity("section")
    properties: List[FullProperty]

    @classmethod
    def from_path(cls, path: Path) -> List[MetaDataClass]:
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist")
        if os.path.isdir(path):
            if os.path.exists(path / ".v7" / "metadata.json"):
                path = path / ".v7" / "metadata.json"
            else:
                raise FileNotFoundError("File metadata.json does not exist in path")
        if path.suffix != ".json":
            raise ValueError(f"File {path} must be a json file")
        with open(path, "r") as f:
            data = json.load(f)
        return [cls.model_validate(d) for d in data["classes"]]


class SelectedProperty(DefaultDarwin):
    """
    Selected property for an annotation found inside a darwin annotation

    Attributes:
        frame_index (int | str): Frame index of the annotation
        int for section-level properties, and "global" for annotation-level properties
        name (str): Name of the property
        type (str | None): Type of the property (if it exists)
        value (str): Value of the property
    """

    frame_index: Optional[Union[int, str]] = None
    name: str
    type: Optional[str] = None
    value: Optional[str] = None



from pathlib import Path

import pytest

from darwin.future.data_objects.properties import MetaDataClass, SelectedProperty


@pytest.fixture
def path_to_metadata_folder() -> Path:
    return Path("darwin/future/tests/data")


@pytest.fixture
def path_to_metadata(path_to_metadata_folder: Path) -> Path:
    return path_to_metadata_folder / ".v7" / "metadata.json"


def test_properties_metadata_loads_folder(path_to_metadata: Path) -> None:
    metadata = MetaDataClass.from_path(path_to_metadata)
    assert metadata is not None
    assert len(metadata) == 2


def test_properties_metadata_loads_file(path_to_metadata: Path) -> None:
    metadata = MetaDataClass.from_path(path_to_metadata)
    assert metadata is not None
    assert len(metadata) == 2


def test_properties_metadata_fails() -> None:
    path = Path("darwin/future/tests/data/does_not_exist.json")
    with pytest.raises(FileNotFoundError):
        MetaDataClass.from_path(path)

    path = Path("darwin/future/tests/data/does_not_exist")
    with pytest.raises(FileNotFoundError):
        MetaDataClass.from_path(path)


def test_can_parse_unpopulated_required_properties() -> None:
    selected_property = SelectedProperty(
        frame_index=None, name="name", type="type", value=None
    )
    assert selected_property is not None



from typing import Literal, Optional

from pydantic import BaseModel, Field, model_validator


class SortingMethods(BaseModel):
    accuracy: Optional[Literal["asc", "desc"]] = Field(None)
    byte_size: Optional[Literal["asc", "desc"]] = Field(None)
    id: Optional[Literal["asc", "desc"]] = Field(None)
    map: Optional[Literal["asc", "desc"]] = Field(None)
    name: Optional[Literal["asc", "desc"]] = Field(None)
    priority: Optional[Literal["asc", "desc"]] = Field(None)
    updated_at: Optional[Literal["asc", "desc"]] = Field(None)

    @model_validator(mode="before")
    def check_at_least_one_field(cls, values):
        assert any(value is not None for value in values.values())
        return values



from __future__ import annotations

from typing import Dict, List, Optional, Protocol, Union, cast
from uuid import UUID

from darwin.future.core.items.archive_items import archive_list_of_items
from darwin.future.core.items.assign_items import assign_items
from darwin.future.core.items.delete_items import delete_list_of_items
from darwin.future.core.items.move_items_to_folder import move_list_of_items_to_folder
from darwin.future.core.items.restore_items import restore_list_of_items
from darwin.future.core.items.set_item_layout import set_item_layout
from darwin.future.core.items.set_item_priority import set_item_priority
from darwin.future.core.items.set_stage_to_items import set_stage_to_items
from darwin.future.core.items.tag_items import tag_items
from darwin.future.core.items.untag_items import untag_items
from darwin.future.data_objects.item import ItemCore, ItemLayout, ItemSlot
from darwin.future.data_objects.workflow import WFStageCore
from darwin.future.exceptions import BadRequest
from darwin.future.meta.objects.base import MetaBase


class hasStage(Protocol):
    # Using Protocol to avoid circular imports between item.py and stage.py
    _element: WFStageCore


class Item(MetaBase[ItemCore]):
    """
    Represents an item in a Darwin dataset.

    Args:
        MetaBase (Stage): Generic MetaBase object expanded by ItemCore object
            return type

    Attributes:
        name (str): The name of the item.
        id (UUID): The unique identifier of the item.
        slots (List[ItemSlot]): A list of slots associated with the item.
        path (str): The path of the item.
        dataset_id (int): The ID of the dataset the item belongs to.
        processing_status (str): The processing status of the item.
        archived (Optional[bool]): Whether the item is archived or not.
        priority (Optional[int]): The priority of the item.
        tags (Optional[Union[List[str], Dict[str, str]]]): The tags associated with the item.
        layout (Optional[ItemLayout]): The layout of the item.

    Methods:
        delete(self) -> None:
            Deletes the item from the Darwin dataset.

    Example usage:
        # Get the item object
        items = workflow.items.where(name='test').collect() # gets first page of items

        # Delete the items
        [item.delete() for item in items] # will collect all pages of items and delete individually

    """

    def delete(self) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        delete_list_of_items(self.client, team_slug, dataset_id, filters)

    def move_to_folder(self, path: str) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        move_list_of_items_to_folder(self.client, team_slug, dataset_id, path, filters)

    def set_priority(self, priority: int) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        set_item_priority(self.client, team_slug, dataset_id, priority, filters)

    def restore(self) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        restore_list_of_items(self.client, team_slug, dataset_id, filters)

    def archive(self) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        archive_list_of_items(self.client, team_slug, dataset_id, filters)

    def set_layout(self, layout: ItemLayout) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        assert isinstance(layout, ItemLayout)
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        set_item_layout(self.client, team_slug, dataset_id, layout, filters)

    def assign(self, assignee_id: int, workflow_id: str | None = None) -> None:
        if not assignee_id:
            raise ValueError("Must specify assignee to assign items to")
        if not workflow_id:
            # if workflow_id is not specified, get it from the meta_params
            # this will be present in the case of a workflow object
            if "workflow_id" in self.meta_params:
                workflow_id = str(self.meta_params["workflow_id"])
            else:
                raise ValueError("Must specify workflow_id to set items to")
        assert isinstance(workflow_id, str)
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)

        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        assign_items(
            self.client, team_slug, dataset_id, assignee_id, workflow_id, filters
        )

    def tag(self, tag_id: int) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        if not isinstance(tag_id, int):
            raise BadRequest(f"tag_id must be an integer, got {type(tag_id)}")
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        tag_items(self.client, team_slug, dataset_id, tag_id, filters)

    def untag(self, tag_id: int) -> None:
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)
        if not isinstance(tag_id, int):
            raise BadRequest(f"tag_id must be an integer, got {type(tag_id)}")
        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        untag_items(self.client, team_slug, dataset_id, tag_id, filters)

    def set_stage(
        self, stage_or_stage_id: hasStage | str, workflow_id: str | None = None
    ) -> None:
        if not stage_or_stage_id:
            raise ValueError(
                "Must specify stage (either Stage object or stage_id string) to set items to"
            )
        if not workflow_id:
            # if workflow_id is not specified, get it from the meta_params
            # this will be present in the case of a workflow object
            if "workflow_id" in self.meta_params:
                workflow_id = str(self.meta_params["workflow_id"])
            else:
                raise ValueError("Must specify workflow_id to set items to")
        assert isinstance(workflow_id, str)
        team_slug, dataset_id = (
            self.meta_params["team_slug"],
            (
                self.meta_params["dataset_id"]
                if "dataset_id" in self.meta_params
                else self.meta_params["dataset_ids"]
            ),
        )
        assert isinstance(team_slug, str)

        # get stage_id from stage_or_stage_id
        if isinstance(stage_or_stage_id, str):
            stage_id = stage_or_stage_id
        else:
            stage_id = str(stage_or_stage_id._element.id)

        dataset_id = cast(Union[int, List[int]], dataset_id)
        filters = {"item_ids": [str(self.id)]}
        set_stage_to_items(
            self.client, team_slug, dataset_id, stage_id, workflow_id, filters
        )

    @property
    def name(self) -> str:
        return self._element.name

    @property
    def id(self) -> UUID:
        return self._element.id

    @property
    def slots(self) -> List[ItemSlot]:
        return self._element.slots

    @property
    def path(self) -> str:
        return self._element.path

    @property
    def dataset_id(self) -> int:
        return self._element.dataset_id

    @property
    def processing_status(self) -> str:
        return self._element.processing_status

    @property
    def archived(self) -> Optional[bool]:
        return self._element.archived

    @property
    def priority(self) -> Optional[int]:
        return self._element.priority

    @property
    def tags(self) -> Optional[Union[List[str], Dict[str, str]]]:
        return self._element.tags

    @property
    def layout(self) -> Optional[ItemLayout]:
        return self._element.layout

    def __str__(self) -> str:
        return f"Item\n\
- Item Name: {self._element.name}\n\
- Item Processing Status: {self._element.processing_status}\n\
- Item ID: {self._element.id}"



import asyncio
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Tuple, Union

from darwin.future.core.client import ClientCore
from darwin.future.data_objects.item import UploadItem
from darwin.future.data_objects.typing import UnknownType
from darwin.future.exceptions import DarwinException

logger = getLogger(__name__)


async def _build_slots(item: UploadItem) -> List[Dict]:
    """
    (internal) Builds the slots for an item

    Parameters
    ----------
    item: UploadItem
        The item to build slots for

    Returns
    -------
    List[Dict]
        The built slots
    """

    if not item.slots:
        return []

    slots_to_return: List[Dict] = []

    for slot in item.slots:
        slot_dict: Dict[str, UnknownType] = {
            "slot_name": slot.slot_name,
            "file_name": slot.file_name,
            "fps": slot.fps,
        }

        if slot.storage_key is not None:
            slot_dict["storage_key"] = slot.storage_key

        if slot.as_frames is not None:
            slot_dict["as_frames"] = slot.as_frames

        if slot.extract_views is not None:
            slot_dict["extract_views"] = slot.extract_views

        if slot.metadata is not None:
            slot_dict["metadata"] = slot.metadata

        if slot.tags is not None:
            slot_dict["tags"] = slot.tags

        if slot.type is not None:
            slot_dict["type"] = slot.type

        slots_to_return.append(slot_dict)

    return slots_to_return


async def _build_layout(item: UploadItem) -> Dict:
    if not item.layout:
        return {}

    if item.layout.version == 1:
        return {
            "slots": item.layout.slots,
            "type": item.layout.type,
            "version": item.layout.version,
        }

    if item.layout.version == 2:
        return {
            "slots": item.layout.slots,
            "type": item.layout.type,
            "version": item.layout.version,
            "layout_shape": item.layout.layout_shape,
        }

    raise DarwinException(f"Invalid layout version {item.layout.version}")


async def _build_payload_items(
    items_and_paths: List[Tuple[UploadItem, Path]],
) -> List[Dict]:
    """
    Builds the payload for the items to be registered for upload

    Parameters
    ----------
    items_and_paths: List[Tuple[UploadItem, Path]]

    Returns
    -------
    List[Dict]
        The payload for the items to be registered for upload
    """

    return_list = []
    for item, path in items_and_paths:
        base_item = {
            "name": getattr(item, "name"),
            "slots": await _build_slots(item),
            "path": str(path),
        }

        if getattr(item, "tags", None):
            base_item["tags"] = item.tags

        if getattr(item, "layout", None):
            base_item["layout"] = await _build_layout(item)

        return_list.append(base_item)

    return return_list


async def async_register_upload(
    api_client: ClientCore,
    team_slug: str,
    dataset_slug: str,
    items_and_paths: Union[Tuple[UploadItem, Path], List[Tuple[UploadItem, Path]]],
    force_tiling: bool = False,
    handle_as_slices: bool = False,
    ignore_dicom_layout: bool = False,
) -> Dict:
    """
    Registers an upload for a dataset that can then be used to upload files to Darwin

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    dataset_slug: str
        The slug of the dataset to register the upload for
    items_and_paths: Union[Tuple[UploadItem, Path], List[Tuple[UploadItem, Path]]]
        A list of tuples of Items and Paths to register for upload
    force_tiling: bool
        Whether to force tiling for the upload
    handle_as_slices: bool
        Whether to handle the upload as slices
    ignore_dicom_layout: bool
        Whether to ignore the dicom layout
    """

    if isinstance(items_and_paths, tuple):
        items_and_paths = [items_and_paths]
        assert all(
            (isinstance(item, UploadItem) and isinstance(path, Path))
            for item, path in items_and_paths
        ), "items must be a list of Items"

    payload_items = await _build_payload_items(items_and_paths)

    options = {
        "force_tiling": force_tiling,
        "handle_as_slices": handle_as_slices,
        "ignore_dicom_layout": ignore_dicom_layout,
    }

    payload = {
        "dataset_slug": dataset_slug,
        "items": payload_items,
        "options": options,
    }

    try:
        response = api_client.post(
            f"/v2/teams/{team_slug}/items/register_upload", payload
        )
    except Exception as exc:
        logger.error(f"Failed to register upload in {__name__}", exc_info=exc)
        raise DarwinException(f"Failed to register upload in {__name__}") from exc

    assert isinstance(response, dict), "Unexpected return type from register upload"

    return response


async def async_create_signed_upload_url(
    api_client: ClientCore,
    team_slug: str,
    upload_id: str,
) -> str:
    """
    Asynchronously create a signed upload URL for an upload or uploads

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    upload_id: str
        The ID of the upload to confirm

    Returns
    -------
    JSONType
        The response from the API
    """
    try:
        response = api_client.get(
            f"/v2/teams/{team_slug}/items/uploads/{upload_id}/sign"
        )
    except Exception as exc:
        logger.error(f"Failed to create signed upload url in {__name__}", exc_info=exc)
        raise DarwinException(
            f"Failed to create signed upload url in {__name__}"
        ) from exc

    assert isinstance(
        response, dict
    ), "Unexpected return type from create signed upload url"

    if not response:
        logger.error(
            f"Failed to create signed upload url in {__name__}, got no response"
        )
        raise DarwinException(
            f"Failed to create signed upload url in {__name__}, got no response"
        )

    if "errors" in response:
        logger.error(
            f"Failed to create signed upload url in {__name__}, got errors: {response['errors']}"
        )
        raise DarwinException(f"Failed to create signed upload url in {__name__}")

    if "upload_url" not in response:
        logger.error(
            f"Failed to create signed upload url in {__name__}, got no upload_url"
        )
        raise DarwinException(
            f"Failed to create signed upload url in {__name__}, got no upload_url"
        )

    return response["upload_url"]


async def async_register_and_create_signed_upload_url(
    api_client: ClientCore,
    team_slug: str,
    dataset_slug: str,
    items_and_paths: Union[Tuple[UploadItem, Path], List[Tuple[UploadItem, Path]]],
    force_tiling: bool = False,
    handle_as_slices: bool = False,
    ignore_dicom_layout: bool = False,
) -> List[Tuple[str, str]]:
    """
    Asynchronously register and create a signed upload URL for an upload or uploads

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    dataset_slug: str
        The slug of the dataset to register the upload for
    items_and_paths: Union[Tuple[Item, Path], List[Tuple[Item, Path]]]
        A list of tuples, or a single tuple of Items and Paths to register for upload
    force_tiling: bool
        Whether to force tiling for the upload
    handle_as_slices: bool
        Whether to handle the upload as slices
    ignore_dicom_layout: bool
        Whether to ignore the dicom layout

    Returns
    -------
    List[Tuple[str, str]]
        List of tuples of signed upload urls and upload ids
    """

    register = await async_register_upload(
        api_client,
        team_slug,
        dataset_slug,
        items_and_paths,
        force_tiling,
        handle_as_slices,
        ignore_dicom_layout,
    )

    if "errors" in register:
        raise DarwinException(f"Failed to register upload in {__name__}")

    if (
        "blocked_items" in register
        and isinstance(register["blocked_items"], list)
        and len(register["blocked_items"]) > 0
    ):
        raise DarwinException(
            f"Failed to register upload in {__name__}, got blocked items: {register['blocked_items']}"
        )

    assert "items" in register, "Unexpected return type from register upload"
    assert "blocked_items" in register, "Unexpected return type from register upload"

    uploaded_items = register["items"]

    upload_ids = []
    for item in uploaded_items:
        if "slots" in item:
            for slot in item["slots"]:
                if "upload_id" in slot:
                    upload_ids.append(slot["upload_id"])

    return [
        (await async_create_signed_upload_url(api_client, team_slug, id), id)
        for id in upload_ids
    ]


async def async_confirm_upload(
    api_client: ClientCore, team_slug: str, upload_id: str
) -> None:
    """
    Asynchronously confirm an upload/uploads was successful by ID

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to confirm the upload for
    upload_id: str
        The ID of the upload to confirm

    Returns
    -------
    JSONType
        The response from the API
    """

    try:
        response = api_client.post(
            f"/v2/teams/{team_slug}/items/uploads/{upload_id}/confirm", data={}
        )
    except Exception as exc:
        logger.error(f"Failed to confirm upload in {__name__}", exc_info=exc)
        raise DarwinException(f"Failed to confirm upload in {__name__}") from exc

    assert isinstance(response, dict), "Unexpected return type from confirm upload"

    if "errors" in response:
        logger.error(
            f"Failed to confirm upload in {__name__}, got errors: {response['errors']}"
        )
        raise DarwinException(
            f"Failed to confirm upload in {__name__}: {str(response['errors'])}"
        )


def register_upload(
    api_client: ClientCore,
    team_slug: str,
    dataset_slug: str,
    items_and_paths: Union[Tuple[UploadItem, Path], List[Tuple[UploadItem, Path]]],
    force_tiling: bool = False,
    handle_as_slices: bool = False,
    ignore_dicom_layout: bool = False,
) -> Dict:
    """
    Asynchronously register an upload/uploads for a dataset that can then be used to upload files to Darwin

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    dataset_slug: str
        The slug of the dataset to register the upload for
    items_and_paths: Union[Tuple[Item, Path], List[Tuple[Item, Path]]]
        A list of tuples, or a single tuple of Items and Paths to register for upload
    force_tiling: bool
        Whether to force tiling for the upload
    handle_as_slices: bool
        Whether to handle the upload as slices
    ignore_dicom_layout: bool
        Whether to ignore the dicom layout
    """

    response = asyncio.run(
        async_register_upload(
            api_client,
            team_slug,
            dataset_slug,
            items_and_paths,
            force_tiling,
            handle_as_slices,
            ignore_dicom_layout,
        )
    )
    return response


def create_signed_upload_url(
    api_client: ClientCore,
    upload_id: str,
    team_slug: str,
) -> str:
    """
    Create a signed upload URL for an upload or uploads

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    upload_id: str
        The ID of the upload to confirm

    Returns
    -------
    JSONType
        The response from the API
    """

    return asyncio.run(async_create_signed_upload_url(api_client, upload_id, team_slug))


def register_and_create_signed_upload_url(
    api_client: ClientCore,
    team_slug: str,
    dataset_slug: str,
    items_and_paths: Union[List[Tuple[UploadItem, Path]], Tuple[UploadItem, Path]],
    force_tiling: bool = False,
    handle_as_slices: bool = False,
    ignore_dicom_layout: bool = False,
) -> List[Tuple[str, str]]:
    """
    Register and create a signed upload URL for an upload or uploads

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to register the upload for
    dataset_slug: str
        The slug of the dataset to register the upload for

    Returns
    -------
    JSONType
        The response from the API
    """

    return asyncio.run(
        async_register_and_create_signed_upload_url(
            api_client,
            team_slug,
            dataset_slug,
            items_and_paths,
            force_tiling,
            handle_as_slices,
            ignore_dicom_layout,
        )
    )


def confirm_upload(api_client: ClientCore, team_slug: str, upload_id: str) -> None:
    """
    Confirm an upload/uploads was successful by ID

    Parameters
    ----------
    api_client: ClientCore
        The client to use for the request
    team_slug: str
        The slug of the team to confirm the upload for
    upload_id: str
        The ID of the upload to confirm

    Returns
    -------
        None

    Raises
    ------
    DarwinException
        If the upload could not be confirmed
    """

    response = asyncio.run(async_confirm_upload(api_client, team_slug, upload_id))
    return response



from typing import List

import pytest
import responses
from responses.matchers import json_params_matcher, query_param_matcher

from darwin.future.core.client import ClientCore
from darwin.future.data_objects.item import ItemLayout
from darwin.future.exceptions import BadRequest
from darwin.future.meta.objects.item import Item
from darwin.future.meta.queries.item import ItemQuery
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.fixtures import *
from darwin.future.tests.meta.objects.fixtures import *


@pytest.fixture
def item_query(base_client: ClientCore) -> ItemQuery:
    return ItemQuery(
        client=base_client, meta_params={"team_slug": "test", "dataset_id": 1}
    )


def test_item_query_collect(item_query: ItemQuery, items_json: List[dict]) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        items = item_query.collect_all()
        assert len(items) == 5
        for i in range(5):
            assert items[i].name == f"item_{i}"


def test_delete(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        rsps.add(
            rsps.DELETE,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item_query.delete()


def test_move_to_folder(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        path = "/new_folder"
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items/path",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "path": path,
                    }
                )
            ],
            json={},
        )
        item_query.move_to_folder(path)


def test_move_to_folder_raises_on_incorrect_parameters(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        path = 1234
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items/path",
            status=400,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "path": path,
                    }
                )
            ],
            json={},
        )
        with pytest.raises(BadRequest):
            item_query.move_to_folder(path)  # type: ignore


def test_set_priority(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        priority = 10
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint
            + f"v2/teams/{team_slug}/items/priority",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "priority": priority,
                    }
                )
            ],
            json={},
        )
        item_query.set_priority(priority)


def test_set_priority_raises_on_incorrect_parameters(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        priority = "10"
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint
            + f"v2/teams/{team_slug}/items/priority",
            status=400,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "priority": priority,
                    }
                )
            ],
            json={},
        )
        with pytest.raises(BadRequest):
            item_query.set_priority(priority)  # type: ignore


def test_restore(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items/restore",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item_query.restore()


def test_archive(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items/archive",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item_query.archive()


def test_set_layout(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        layout = ItemLayout(version=1, type="grid", slots=["slot1", "slot2"])
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint + f"v2/teams/{team_slug}/items/layout",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "layout": dict(layout),
                    }
                )
            ],
            json={},
        )
        item_query.set_layout(layout)


def test_set_layout_raises_on_incorrect_parameters(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock():
        items[0].meta_params["team_slug"]
        items[0].meta_params["dataset_id"]
        layout = "invalid_layout"
        with pytest.raises(AssertionError):
            item_query.set_layout(layout)  # type: ignore


def test_tag(item_query: ItemQuery, items_json: List[dict], items: List[Item]) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        tag_id = 123456
        rsps.add(
            rsps.POST,
            items[0].client.config.api_endpoint
            + f"v2/teams/{team_slug}/items/slots/tags",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "annotation_class_id": tag_id,
                    }
                )
            ],
            json={},
        )
        item_query.tag(tag_id)


def test_tag_bad_request(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock():
        tag_id = "123456"
        with pytest.raises(BadRequest) as excinfo:
            item_query.tag(tag_id)  # type: ignore
        (msg,) = excinfo.value.args
        assert msg == "tag_id must be an integer, got <class 'str'>"


def test_untag(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            item_query.client.config.api_endpoint + "v2/teams/test/items",
            match=[
                query_param_matcher(
                    {"page[offset]": "0", "page[size]": "500", "dataset_ids": "1"}
                )
            ],
            json={"items": items_json, "errors": []},
        )
        team_slug = items[0].meta_params["team_slug"]
        dataset_id = items[0].meta_params["dataset_id"]
        tag_id = 123456
        rsps.add(
            rsps.DELETE,
            items[0].client.config.api_endpoint
            + f"v2/teams/{team_slug}/items/slots/tags",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id) for item in items],
                            "dataset_ids": [dataset_id],
                        },
                        "annotation_class_id": tag_id,
                    }
                )
            ],
            json={},
        )
        item_query.untag(tag_id)


def test_untag_bad_request(
    item_query: ItemQuery, items_json: List[dict], items: List[Item]
) -> None:
    with responses.RequestsMock():
        tag_id = "123456"
        with pytest.raises(BadRequest) as excinfo:
            item_query.untag(tag_id)  # type: ignore
        (msg,) = excinfo.value.args
        assert msg == "tag_id must be an integer, got <class 'str'>"


def test_sort_method(base_client: ClientCore):
    item_query = ItemQuery(
        base_client, meta_params={"dataset_id": 0000, "team_slug": "test_team"}
    )

    item_query.sort(accuracy="desc", byte_size="asc")

    assert len(item_query.filters) == 2
    assert item_query.filters[0].name == "sort[accuracy]"
    assert item_query.filters[0].param == "desc"



from typing import Optional, Union

from darwin.future.core.client import ClientCore
from darwin.future.core.datasets import get_dataset, remove_dataset
from darwin.future.core.team.get_team import get_team
from darwin.future.data_objects.team import TeamCore
from darwin.future.exceptions import MissingDataset
from darwin.future.helpers.assertion import assert_is
from darwin.future.meta.objects.base import MetaBase
from darwin.future.meta.objects.dataset import Dataset
from darwin.future.meta.queries.dataset import DatasetQuery
from darwin.future.meta.queries.item import ItemQuery
from darwin.future.meta.queries.team_member import TeamMemberQuery
from darwin.future.meta.queries.workflow import WorkflowQuery


class Team(MetaBase[TeamCore]):
    """
    Team Meta object. Facilitates the creation of Query objects, lazy loading of sub
    fields like members unlike other MetaBase objects, does not extend the __next__
    function because it is not iterable. This is because Team is linked to api key and
    only one team can be returned, but stores a list of teams for consistency. This
    does mean however that to access the underlying team object, you must access the
    first element of the list
    team = client.team[0]

    Args:
        MetaBase (Team): Generic MetaBase object expanded by Team core object return
            type

    Returns:
        Team: Team object

    Attributes:
        name (str): The name of the team.
        slug (str): The slug of the team.
        id (int): The id of the team.
        members (TeamMemberQuery): A query of team members associated with the team.
        datasets (DatasetQuery): A query of datasets associated with the team.
        workflows (WorkflowQuery): A query of workflows associated with the team.

    Methods:
        create_dataset(slug: str) -> Dataset:
            Creates a new dataset with the given name and slug.
        delete_dataset(client: Client, id: int) -> None:
            Removes the dataset with the given slug.


    Example Usage:
        # Get the team object
        client = Client.local()
        team = client.team[0]

        # Get a dataset object associated with the team
        dataset = team.datasets.where(name="my_dataset_name").collect_one()

        # Create a new dataset associated with the team
        new_dataset = team.create_dataset(name="new_dataset", slug="new_dataset_slug")

        # Remove a dataset associated with the team
        team.remove_dataset(slug="my_dataset_slug")

        # Get a workflow object associated with the team
        workflow = team.workflows.where(name="my_workflow_name").collect_one()

        # Get a team member object associated with the team
        team_member = team.members.where(email="...")
    """

    def __init__(self, client: ClientCore, team: Optional[TeamCore] = None) -> None:
        team = team or get_team(client)
        super().__init__(client=client, element=team)

    @property
    def name(self) -> str:
        return self._element.name

    @property
    def id(self) -> int:
        assert self._element.id is not None
        return self._element.id

    @property
    def members(self) -> TeamMemberQuery:
        return TeamMemberQuery(self.client, meta_params={"team_slug": self.slug})

    @property
    def slug(self) -> str:
        return self._element.slug

    @property
    def datasets(self) -> DatasetQuery:
        return DatasetQuery(self.client, meta_params={"team_slug": self.slug})

    @property
    def workflows(self) -> WorkflowQuery:
        return WorkflowQuery(self.client, meta_params={"team_slug": self.slug})

    @property
    def items(self) -> ItemQuery:
        return ItemQuery(
            self.client, meta_params={"team_slug": self.slug, "dataset_ids": "all"}
        )

    @classmethod
    def delete_dataset(cls, client: ClientCore, dataset_id: Union[int, str]) -> int:
        """
        Deletes a dataset by id or slug

        Parameters
        ----------
        dataset_id: Union[int, str]
            The id or slug of the dataset to delete

        Returns
        -------
        Tuple[Optional[List[Exception]], int]
            A tuple containing a list of exceptions and the number of datasets deleted
        """
        if isinstance(dataset_id, str):
            dataset_deleted = cls._delete_dataset_by_slug(client, dataset_id)
        else:
            dataset_deleted = cls._delete_dataset_by_id(client, dataset_id)

        return dataset_deleted

    @staticmethod
    def _delete_dataset_by_slug(client: ClientCore, slug: str) -> int:
        """
        (internal) Deletes a dataset by slug

        Parameters
        ----------
        client: MetaClient
            The client to use to make the request

        slug: str
            The slug of the dataset to delete

        Returns
        -------
        int
            The dataset deleted
        """
        assert_is(isinstance(client, ClientCore), "client must be a Core Client")
        assert_is(isinstance(slug, str), "slug must be a string")

        dataset = get_dataset(client, slug)
        if dataset and dataset.id:
            dataset_deleted = remove_dataset(client, dataset.id)
        else:
            raise MissingDataset(f"Dataset with slug {slug} not found")

        return dataset_deleted

    @staticmethod
    def _delete_dataset_by_id(client: ClientCore, dataset_id: int) -> int:
        """
        (internal) Deletes a dataset by id

        Parameters
        ----------
        client: Client
            The client to use to make the request

        dataset_id: int
            The id of the dataset to delete

        Returns
        -------
        int
            The dataset deleted
        """
        assert_is(isinstance(client, ClientCore), "client must be a Client")
        assert_is(isinstance(dataset_id, int), "dataset_id must be an integer")

        dataset_deleted = remove_dataset(client, dataset_id)
        return dataset_deleted

    def create_dataset(self, slug: str) -> Dataset:
        core = Dataset.create_dataset(self.client, slug)
        return Dataset(
            client=self.client, element=core, meta_params={"team_slug": self.slug}
        )

    def __str__(self) -> str:
        return f"Team\n\
- Team Name: {self._element.name}\n\
- Team Slug: {self._element.slug}\n\
- Team ID: {self._element.id}\n\
- {len(self._element.members if self._element.members else [])} member(s)"



from __future__ import annotations

from typing import List, Optional, Sequence, Union
from uuid import UUID

from darwin.cli_functions import upload_data
from darwin.dataset.upload_manager import LocalFile
from darwin.datatypes import PathLike
from darwin.future.data_objects.workflow import WFDatasetCore, WFTypeCore, WorkflowCore
from darwin.future.exceptions import MissingDataset
from darwin.future.meta.objects.base import MetaBase
from darwin.future.meta.queries.item import ItemQuery
from darwin.future.meta.queries.stage import Stage, StageQuery


class Workflow(MetaBase[WorkflowCore]):
    """
    Workflow Meta object. Facilitates the creation of Query objects, lazy loading of
    sub fields

    Args:
        MetaBase (Workflow): Generic MetaBase object expanded by Workflow core object
            return type

    Returns:
        _type_: Workflow

    Attributes:
        name (str): The name of the workflow.
        id (UUID): The id of the workflow
        datasets (List[Dataset]): A list of datasets associated with the workflow.
        stages (StageQuery): Queries stages associated with the workflow.

    Methods:
        push_from_dataset_stage() -> Workflow:
            moves all items associated with the dataset stage to the next connected stage
        upload_files(...): -> Workflow:
            Uploads files to the dataset stage of the workflow

    Example Usage:
        # Get the workflow object
        workflow = client.team.workflows.where(name='test').collect_one()

        # Get the stages associated with the workflow
        stages = workflow.stages

        # Get the datasets associated with the workflow
        datasets = workflow.datasets
    """

    @property
    def items(self) -> ItemQuery:
        meta_params = self.meta_params.copy()
        meta_params["dataset_ids"] = str(self.datasets[0].id)
        meta_params["workflow_id"] = str(self.id)
        return ItemQuery(self.client, meta_params=meta_params)

    @property
    def stages(self) -> StageQuery:
        meta_params = self.meta_params.copy()
        meta_params["workflow_id"] = self._element.id
        if self.datasets is not None:
            meta_params["dataset_id"] = self.datasets[0].id
            meta_params["dataset_name"] = self.datasets[0].name
        return StageQuery(self.client, meta_params=meta_params)

    def _get_dataset_stage(self) -> Stage:
        # stages are not in right order - finding the dataset stage
        for stage in self.stages:
            if stage.type == "dataset":
                return stage

        raise MissingDataset("Workflow has no dataset stage")

    @property
    def datasets(self) -> List[WFDatasetCore]:
        if self._element.dataset is None:
            raise ValueError("WorkflowMeta has no associated dataset")
        return [self._element.dataset]

    @property
    def id(self) -> UUID:
        return self._element.id

    @property
    def name(self) -> str:
        return self._element.name

    def push_from_dataset_stage(
        self, wait: bool = True, wait_max_attempts: int = 5, wait_time: float = 0.5
    ) -> Workflow:
        assert self._element.dataset is not None
        stages = self.stages
        assert len(stages) > 1

        ds_stage = self._get_dataset_stage()
        assert ds_stage._element.type == WFTypeCore.DATASET
        next_stage = ds_stage._element.edges[0].target_stage_id
        assert next_stage is not None
        ds_stage.move_attached_files_to_stage(
            next_stage, wait, wait_max_attempts, wait_time
        )

        return self

    def upload_files(
        self,
        files: Sequence[Union[PathLike, LocalFile]],
        files_to_exclude: Optional[List[PathLike]] = None,
        fps: int = 1,
        path: Optional[str] = None,
        frames: bool = False,
        extract_views: bool = False,
        preserve_folders: bool = False,
        verbose: bool = False,
        auto_push: bool = True,
        wait: bool = True,
        wait_max_attempts: int = 5,
        wait_time: float = 0.5,
    ) -> Workflow:
        assert self._element.dataset is not None
        upload_data(
            self.datasets[0].name,
            files,  # type: ignore
            files_to_exclude,
            fps,
            path,
            frames,
            extract_views,
            preserve_folders,
            verbose,
        )
        if auto_push:
            self.push_from_dataset_stage(
                wait=wait, wait_max_attempts=wait_max_attempts, wait_time=wait_time
            )
        return self

    def __str__(self) -> str:
        return f"Workflow\n\
- Workflow Name: {self._element.name}\n\
- Workflow ID: {self._element.id}\n\
- Connected Dataset ID: {self.datasets[0].id}\n\
- Conneted Dataset Name: {self.datasets[0].name}"



from __future__ import annotations

from typing import List, Optional, Sequence, Union

from darwin.cli_functions import upload_data
from darwin.dataset.upload_manager import LocalFile
from darwin.datatypes import PathLike
from darwin.future.core.client import ClientCore
from darwin.future.core.datasets import create_dataset, remove_dataset
from darwin.future.data_objects.dataset import DatasetCore
from darwin.future.helpers.assertion import assert_is
from darwin.future.meta.objects.base import MetaBase
from darwin.future.meta.queries.item import ItemQuery
from darwin.future.meta.queries.item_id import ItemIDQuery


class Dataset(MetaBase[DatasetCore]):
    """
    Dataset Meta object. Facilitates the management of a dataset, querying of items,
    uploading data, and other dataset related operations.

    Args:
        MetaBase (Dataset): Generic MetaBase object that manages a DatasetCore object

    Returns:
        _type_: DatasetMeta

    Attributes:
        name (str): The name of the dataset.
        slug (str): The slug of the dataset.
        id (int): The id of the dataset.
        item_ids (List[UUID]): A list of item ids for the dataset.

    Example Usage:
        # Create a new dataset
        dataset = Dataset.create(client, slug="my_dataset_slug")

        # Upload data to the dataset
        local_file = LocalFile("path/to/local/file")
        upload_data(dataset, [local_file])

        # Get the item ids for the dataset
        item_ids = dataset.item_ids

        # Remove the dataset
        dataset.remove()
    """

    @property
    def name(self) -> str:
        assert self._element.name is not None
        return self._element.name

    @property
    def slug(self) -> str:
        assert self._element.slug is not None
        return self._element.slug

    @property
    def id(self) -> int:
        assert self._element.id is not None
        return self._element.id

    @property
    def item_ids(self) -> ItemIDQuery:
        """Returns a list of item ids for the dataset

        Returns:
            List[UUID]: A list of item ids
        """
        assert self._element.id is not None
        assert self.meta_params["team_slug"] is not None and isinstance(
            self.meta_params["team_slug"], str
        )
        meta_params = {"dataset_ids": self.id, **self.meta_params}
        return ItemIDQuery(self.client, meta_params=meta_params)

    @property
    def items(self) -> ItemQuery:
        meta_params = {"dataset_ids": self.id, **self.meta_params}
        return ItemQuery(self.client, meta_params=meta_params)

    @classmethod
    def create_dataset(cls, client: ClientCore, slug: str) -> DatasetCore:
        """
        Creates a new dataset for the given team

        Parameters
        ----------
        slug: str [a-b0-9-_]
            The slug of the dataset to create

        Returns
        -------
        Tuple[Optional[List[Exception]], Optional[Dataset]]
            A tuple containing a list of exceptions and the dataset created

        """
        dataset: Optional[DatasetCore] = None
        cls._validate_slug(slug)
        dataset = create_dataset(client, slug)
        return dataset

    def update_dataset(self) -> DatasetCore:
        # TODO: implement in IO-1018
        raise NotImplementedError()

    def delete(self) -> int:
        """
        Deletes a current dataset

        Returns
        -------
        int
            The id of dataset deleted
        """
        dataset_deleted = remove_dataset(self.client, self.id)
        return dataset_deleted

    @staticmethod
    def _validate_slug(slug: str) -> None:
        """
        Validates a slug

        Parameters
        ----------
        slug: str
            The slug to validate

        Raises
        ------
        AssertionError
        """
        slug_copy = str(slug).lower().strip()
        assert_is(isinstance(slug_copy, str), "slug must be a string")
        assert_is(len(slug_copy) > 0, "slug must not be empty")

        VALID_SLUG_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789-_"
        assert_is(
            all(c in VALID_SLUG_CHARS for c in slug_copy),
            "slug must only contain valid characters",
        )

    def upload_files(
        self,
        files: Sequence[Union[PathLike, LocalFile]],
        files_to_exclude: Optional[List[PathLike]] = None,
        fps: int = 1,
        path: Optional[str] = None,
        frames: bool = False,
        extract_views: bool = False,
        preserve_folders: bool = False,
        verbose: bool = False,
    ) -> Dataset:
        upload_data(
            self._element.name,
            files,  # type: ignore
            files_to_exclude,
            fps,
            path,
            frames,
            extract_views,
            preserve_folders,
            verbose,
        )
        return self

    def __str__(self) -> str:
        releases = self._element.releases
        return f"Dataset\n\
- Name: {self._element.name}\n\
- Dataset Slug: {self._element.slug}\n\
- Dataset ID: {self._element.id}\n\
- Dataset Releases: {releases if releases else 'No releases'}"



from pydantic import BaseModel, ConfigDict


class DefaultDarwin(BaseModel):
    """Default Darwin-Py pydantic settings for meta information.
    Default settings include:
        - auto validating variables on setting/assignment
        - underscore attributes are private
        - objects are passed by reference to prevent unnecesary data copying
    """

    model_config = ConfigDict(validate_assignment=True, protected_namespaces=())



import csv
from pathlib import Path
from typing import Dict, List, Optional, Union

import darwin.datatypes as dt


def parse_path(path: Path) -> Optional[List[dt.AnnotationFile]]:
    """
    Parses the given ``csv video`` file and returns a ``List[dt.AnnotationFile]``, or ``None`` if
    the extension of the given file is not ``.csv``.

    Parameters
    ----------
    path : Path
        The ``Path`` to the file to be parsed.

    Returns
    -------
    Optional[List[dt.AnnotationFile]]
        A ``List[dt.AnnotationFile]`` or ``None`` if the file was parseable.
    """
    if path.suffix != ".csv":
        return None

    files = []

    file_annotation_map: Dict[str, List[Union[dt.Annotation, dt.VideoAnnotation]]] = {}
    with path.open() as f:
        reader = csv.reader(f)
        for row in reader:
            try:
                filename, tag, start_frame, end_frame = (s.strip() for s in row)
            except ValueError:
                continue
            if filename == "":
                continue

            start_frame = int(start_frame)
            end_frame = int(end_frame)

            annotation = dt.make_tag(tag)
            frames = {i: annotation for i in range(start_frame, end_frame + 1)}
            keyframes = {i: i == start_frame for i in range(start_frame, end_frame + 1)}

            annotation = dt.make_video_annotation(
                frames, keyframes, [[start_frame, end_frame]], False, slot_names=[]
            )
            if filename not in file_annotation_map:
                file_annotation_map[filename] = []
            file_annotation_map[filename].append(annotation)
    for filename in file_annotation_map:
        annotations = file_annotation_map[filename]
        annotation_classes = {annotation.annotation_class for annotation in annotations}
        filename_path = Path(filename)
        remote_path = str(filename_path.parent)
        if not remote_path.startswith("/"):
            remote_path = "/" + remote_path
        filename = filename_path.name
        files.append(
            dt.AnnotationFile(
                path,
                filename,
                annotation_classes,
                annotations,
                is_video=True,
                remote_path=remote_path,
            )
        )
    return files



from darwin.future.data_objects.team import TeamMemberCore
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.meta.objects.base import MetaBase


class TeamMember(MetaBase[TeamMemberCore]):
    """
    Team Member Meta object. Facilitates the creation of Query objects, lazy loading of
    sub fields

    Args:
        MetaBase (TeamMember): Generic MetaBase object expanded by TeamMemberCore object
            return type

    Returns:
        _type_: TeamMember

    Attributes:
        first_name (str): The first name of the team member.
        last_name (str): The last name of the team member.
        email (str): The email of the team member.
        user_id (int): The user id of the team member.
        role (TeamMemberRole): The role of the team member.

    Methods:
        None

    Example Usage:
        # Get the role of the team member
        team_member = client.team.members
            .where(first_name='John', last_name='Doe')
            .collect_one()

        role = team_member.role
    """

    @property
    def role(self) -> TeamMemberRole:
        return self._element.role

    @property
    def first_name(self) -> str:
        return self._element.first_name

    @property
    def last_name(self) -> str:
        return self._element.last_name

    @property
    def email(self) -> str:
        return self._element.email

    @property
    def user_id(self) -> int:
        return self._element.user_id

    def __str__(self) -> str:
        return f"Team Member\n\
- Name: {self.first_name} {self.last_name}\n\
- Role: {self.role.value}\n\
- Email: {self.email}\n\
- User ID: {self.user_id}"



import asyncio
from pathlib import Path
from typing import Dict, Generator, List, Tuple
from unittest.mock import MagicMock, Mock, patch

import orjson as json
import pytest
import responses

import darwin.future.core.items.uploads as uploads
from darwin.future.core.client import ClientCore, DarwinConfig
from darwin.future.data_objects.item import ItemLayout, ItemSlot, UploadItem
from darwin.future.exceptions import DarwinException
from darwin.future.tests.core.fixtures import *  # noqa: F401,F403

from .fixtures import *  # noqa: F401,F403


class TestBuildSlots:
    BUILD_SLOT_RETURN_TYPE = List[Dict]

    items_and_expectations: List[Tuple[UploadItem, BUILD_SLOT_RETURN_TYPE]] = []

    # Test empty slots
    items_and_expectations.append(
        (
            UploadItem(name="name_with_no_slots", slots=[]),
            [],
        )
    )

    # Test Simple slot with no non-required fields
    items_and_expectations.append(
        (
            UploadItem(
                name="name_with_simple_slot",
                slots=[
                    ItemSlot(
                        slot_name="slot_name_simple",
                        file_name="file_name",
                        storage_key="storage_key",
                    )
                ],
            ),
            [
                {
                    "slot_name": "slot_name_simple",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                }
            ],
        )
    )

    # Test with multiple slots
    items_and_expectations.append(
        (
            UploadItem(
                name="name_with_multiple_slots",
                slots=[
                    ItemSlot(
                        slot_name="slot_name1",
                        file_name="file_name1",
                        storage_key="storage_key1",
                    ),
                    ItemSlot(
                        slot_name="slot_name2",
                        file_name="file_name2",
                        storage_key="storage_key2",
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name1",
                    "file_name": "file_name1",
                    "storage_key": "storage_key1",
                    "fps": 0,
                },
                {
                    "slot_name": "slot_name2",
                    "file_name": "file_name2",
                    "storage_key": "storage_key2",
                    "fps": 0,
                },
            ],
        )
    )

    # Test with `as_frames` optional field
    items_and_expectations.append(
        (
            UploadItem(
                name="name_testing_as_frames",
                slots=[
                    ItemSlot(
                        slot_name="slot_name1",
                        file_name="file_name",
                        storage_key="storage_key",
                        as_frames=True,
                    ),
                    ItemSlot(
                        slot_name="slot_name2",
                        file_name="file_name",
                        storage_key="storage_key",
                        as_frames=False,
                    ),
                    ItemSlot(
                        slot_name="slot_name3",
                        file_name="file_name",
                        storage_key="storage_key",
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name1",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                    "as_frames": True,
                },
                {
                    "slot_name": "slot_name2",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                    "as_frames": False,
                },
                {
                    "slot_name": "slot_name3",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                },
            ],
        )
    )

    # Test with `extract_views` optional field
    items_and_expectations.append(
        (
            UploadItem(
                name="name_testing_extract_views",
                slots=[
                    ItemSlot(
                        slot_name="slot_name1",
                        file_name="file_name",
                        storage_key="storage_key",
                        extract_views=True,
                    ),
                    ItemSlot(
                        slot_name="slot_name2",
                        file_name="file_name",
                        storage_key="storage_key",
                        extract_views=False,
                    ),
                    ItemSlot(
                        slot_name="slot_name3",
                        file_name="file_name",
                        storage_key="storage_key",
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name1",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                    "extract_views": True,
                },
                {
                    "slot_name": "slot_name2",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                    "extract_views": False,
                },
                {
                    "slot_name": "slot_name3",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                },
            ],
        )
    )

    # Test with `fps` semi-optional field - field defaults to 0 if not provided
    items_and_expectations.append(
        (
            UploadItem(
                name="name_with_simple_slot",
                slots=[
                    ItemSlot(
                        slot_name="slot_name25",
                        file_name="file_name",
                        storage_key="storage_key",
                        fps=25,  # Testing int
                    ),
                    ItemSlot(
                        slot_name="slot_name29.997",
                        file_name="file_name",
                        storage_key="storage_key",
                        fps=29.997,  # Testing float
                    ),
                    ItemSlot(
                        slot_name="slot_namenative",
                        file_name="file_name",
                        storage_key="storage_key",
                        fps="native",  # Testing literal
                    ),
                    ItemSlot(
                        slot_name="slot_name",
                        file_name="file_name",
                        storage_key="storage_key",
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name25",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 25,
                },
                {
                    "slot_name": "slot_name29.997",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 29.997,
                },
                {
                    "slot_name": "slot_namenative",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": "native",
                },
                {
                    "slot_name": "slot_name",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                },
            ],
        )
    )

    # Test with `metadata` optional field
    items_and_expectations.append(
        (
            UploadItem(
                name="name_with_simple_slot",
                slots=[
                    ItemSlot(
                        slot_name="slot_name",
                        file_name="file_name",
                        storage_key="storage_key",
                        metadata={"key": "value"},
                    ),
                    ItemSlot(
                        slot_name="slot_name",
                        file_name="file_name",
                        storage_key="storage_key",
                        metadata=None,
                    ),
                    ItemSlot(
                        slot_name="slot_name",
                        file_name="file_name",
                        storage_key="storage_key",
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                    "metadata": {"key": "value"},
                },
                {
                    "slot_name": "slot_name",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                },
                {
                    "slot_name": "slot_name",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "fps": 0,
                },
            ],
        )
    )

    # Test with `tags` optional field
    items_and_expectations.append(
        (
            UploadItem(
                name="name_testing_tags",
                slots=[
                    ItemSlot(
                        slot_name="slot_name_with_string_list",
                        file_name="file_name",
                        storage_key="storage_key",
                        tags=["tag1", "tag2"],
                    ),
                    ItemSlot(
                        slot_name="slot_name_with_kv_pairs",
                        file_name="file_name",
                        storage_key="storage_key",
                        tags={"key": "value"},
                    ),
                ],
            ),
            [
                {
                    "slot_name": "slot_name_with_string_list",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "tags": ["tag1", "tag2"],
                    "fps": 0,
                },
                {
                    "slot_name": "slot_name_with_kv_pairs",
                    "file_name": "file_name",
                    "storage_key": "storage_key",
                    "tags": {"key": "value"},
                    "fps": 0,
                },
            ],
        )
    )

    @pytest.mark.parametrize("item,expected", items_and_expectations)
    def test_build_slots(self, item: UploadItem, expected: List[Dict]) -> None:
        result = asyncio.run(uploads._build_slots(item))
        assert result == expected


class TestBuildLayout:
    @pytest.mark.parametrize(
        "item, expected",
        [
            (
                UploadItem(
                    name="test_item",
                    layout=ItemLayout(version=1, type="grid", slots=["slot1", "slot2"]),
                ),
                {
                    "slots": ["slot1", "slot2"],
                    "type": "grid",
                    "version": 1,
                },
            ),
            (
                UploadItem(
                    name="test_item",
                    layout=ItemLayout(
                        version=2,
                        type="grid",
                        slots=["slot1", "slot2"],
                        layout_shape=[3, 4],
                    ),
                ),
                {
                    "slots": ["slot1", "slot2"],
                    "type": "grid",
                    "version": 2,
                    "layout_shape": [3, 4],
                },
            ),
        ],
    )
    def test_build_layout(self, item: UploadItem, expected: Dict) -> None:
        assert asyncio.run(uploads._build_layout(item)) == expected


class TestBuildPayloadItems:
    @pytest.mark.parametrize(
        "items_and_paths, expected",
        [
            (
                [
                    (
                        UploadItem(
                            name="test_item",
                            slots=[
                                ItemSlot(
                                    slot_name="slot_name_with_string_list",
                                    file_name="file_name",
                                    storage_key="storage_key",
                                    tags=["tag1", "tag2"],
                                ),
                                ItemSlot(
                                    slot_name="slot_name_with_kv_pairs",
                                    file_name="file_name",
                                    storage_key="storage_key",
                                    tags={"key": "value"},
                                ),
                            ],
                        ),
                        Path("test_path"),
                    )
                ],
                [
                    {
                        "name": "test_item",
                        "path": "test_path",
                        "slots": [
                            {
                                "slot_name": "slot_name_with_string_list",
                                "file_name": "file_name",
                                "storage_key": "storage_key",
                                "tags": ["tag1", "tag2"],
                                "fps": 0,
                            },
                            {
                                "slot_name": "slot_name_with_kv_pairs",
                                "file_name": "file_name",
                                "storage_key": "storage_key",
                                "tags": {"key": "value"},
                                "fps": 0,
                            },
                        ],
                    }
                ],
            )
        ],
    )
    def test_build_payload_items(
        self, items_and_paths: List[Tuple[UploadItem, Path]], expected: List[Dict]
    ) -> None:
        result = asyncio.run(uploads._build_payload_items(items_and_paths))

        assert result == expected


class SetupTests:
    @pytest.fixture
    def default_url(self, base_client: ClientCore) -> str:
        return f"{base_client.config.base_url}api/v2/teams/my-team/items"


class TestRegisterUpload(SetupTests):
    team_slug = "my-team"
    dataset_slug = "my-dataset"

    @pytest.fixture
    def items_and_paths(self) -> List[Tuple[UploadItem, Path]]:
        return [
            (
                UploadItem(
                    name="test_item",
                    slots=[
                        ItemSlot(
                            slot_name="slot_name_with_string_list",
                            file_name="file_name",
                            storage_key="storage_key",
                            tags=["tag1", "tag2"],
                        ),
                        ItemSlot(
                            slot_name="slot_name_with_kv_pairs",
                            file_name="file_name",
                            storage_key="storage_key",
                            tags={"key": "value"},
                        ),
                    ],
                ),
                Path("test_path"),
            )
        ]

    @responses.activate
    @patch.object(uploads, "_build_payload_items")
    def test_async_register_upload_happy_path(
        self,
        mock_build_payload_items: MagicMock,
        base_client: ClientCore,
        default_url: str,
        items_and_paths: List[Tuple[UploadItem, Path]],
    ) -> None:
        items = [
            {
                "name": "test_item",
                "path": "test_path",
                "tags": [],
                "slots": [
                    {
                        "slot_name": "slot_name_with_string_list",
                        "file_name": "file_name",
                        "storage_key": "storage_key",
                        "tags": ["tag1", "tag2"],
                        "fps": 0,
                        "type": "image",
                    },
                    {
                        "slot_name": "slot_name_with_kv_pairs",
                        "file_name": "file_name",
                        "storage_key": "storage_key",
                        "tags": {"key": "value"},
                        "fps": 0,
                        "type": "image",
                    },
                ],
            }
        ]
        mock_build_payload_items.return_value = items

        responses.add(
            responses.POST, f"{default_url}/register_upload", json={"status": "success"}
        )
        asyncio.run(
            uploads.async_register_upload(
                api_client=base_client,
                team_slug=self.team_slug,
                dataset_slug=self.dataset_slug,
                items_and_paths=items_and_paths,
                force_tiling=True,
                handle_as_slices=True,
                ignore_dicom_layout=True,
            )
        )
        # noqa: E501
        assert responses.calls[0].request.url == f"{default_url}/register_upload"  # type: ignore
        # noqa: E501
        received_call = json.loads(responses.calls[0].request.body.decode("utf-8"))  # type: ignore
        assert received_call == {
            "dataset_slug": "my-dataset",
            "items": items,
            "options": {
                "force_tiling": True,
                "handle_as_slices": True,
                "ignore_dicom_layout": True,
            },
        }, "The request body should be empty"

    @patch.object(uploads, "_build_payload_items")
    def test_async_register_upload_raises(
        self, mock_build_payload_items: MagicMock, base_client: ClientCore
    ) -> None:
        with pytest.raises(DarwinException) as exc:
            mock_build_payload_items.side_effect = DarwinException("Error1")

            asyncio.run(
                uploads.async_register_upload(
                    api_client=base_client,
                    team_slug=self.team_slug,
                    dataset_slug=self.dataset_slug,
                    items_and_paths=[],
                    force_tiling=True,
                    handle_as_slices=True,
                    ignore_dicom_layout=True,
                )
            )

            assert str(exc) == "Error1", "Failed to raise on failed payload build"

        with pytest.raises(DarwinException) as exc:
            base_client.post = MagicMock()  # type: ignore
            base_client.post.side_effect = DarwinException("Error2")

            asyncio.run(
                uploads.async_register_upload(
                    api_client=base_client,
                    team_slug=self.team_slug,
                    dataset_slug=self.dataset_slug,
                    items_and_paths=[],
                    force_tiling=True,
                    handle_as_slices=True,
                    ignore_dicom_layout=True,
                )
            )

            assert str(exc) == "Error2", "Failed to raise on failed API call"


class TestCreateSignedUploadUrl(SetupTests):
    def test_async_create_signed_upload_url(
        self, default_url: str, base_config: DarwinConfig
    ) -> None:
        with responses.RequestsMock() as rsps:
            # Mock the API response
            expected_response = {"upload_url": "https://signed.url"}
            rsps.add(
                rsps.GET,
                f"{default_url}/uploads/1/sign",
                json=expected_response,
            )

            # Call the function with mocked arguments
            api_client = ClientCore(base_config)
            actual_response = asyncio.run(
                uploads.async_create_signed_upload_url(api_client, "my-team", "1")
            )

            # Check that the response matches the expected response
            if not actual_response:
                pytest.fail("Response was None")

            assert actual_response == expected_response["upload_url"]

    def test_async_create_signed_upload_url_raises(
        self, base_client: ClientCore
    ) -> None:
        base_client.post = MagicMock()  # type: ignore
        base_client.post.side_effect = DarwinException("Error")

        with pytest.raises(DarwinException):
            asyncio.run(
                uploads.async_create_signed_upload_url(base_client, "1", "my-team")
            )


class TestRegisterAndCreateSignedUploadUrl:
    @pytest.fixture
    def mock_async_register_upload(self) -> Generator:
        with patch.object(uploads, "async_register_upload") as mock:
            yield mock

    @pytest.fixture
    def mock_async_create_signed_upload_url(self) -> Generator:
        with patch.object(uploads, "async_create_signed_upload_url") as mock:
            yield mock

    def test_async_register_and_create_signed_upload_url(
        self,
        mock_async_register_upload: MagicMock,
        mock_async_create_signed_upload_url: MagicMock,
    ) -> None:
        # Set up mock responses
        mock_async_register_upload.return_value = {
            "blocked_items": [],
            "items": [{"id": "123", "slots": [{"upload_id": "321"}]}],
        }
        mock_signed_url_response = "https://signed.url"
        mock_async_create_signed_upload_url.return_value = mock_signed_url_response

        # Set up mock API client
        mock_api_client = MagicMock()

        # Call the function with mocked arguments
        actual_response = asyncio.run(
            uploads.async_register_and_create_signed_upload_url(
                mock_api_client,
                "my-team",
                "my-dataset",
                [(Mock(), Mock())],
                False,
                False,
                False,
            )
        )

        # Check that the function called the correct sub-functions with the correct arguments
        mock_async_register_upload.assert_called_once()
        mock_async_create_signed_upload_url.assert_called_once_with(
            mock_api_client,
            "my-team",
            "321",
        )

        # Check that the response matches the expected response
        assert actual_response == [("https://signed.url", "321")]

    def test_async_register_and_create_signed_upload_url_raises(
        self,
        mock_async_register_upload: MagicMock,
        mock_async_create_signed_upload_url: MagicMock,
    ) -> None:
        # Set up mock responses
        mock_async_register_upload.return_value = {"id": "123", "errors": ["error"]}
        mock_signed_url_response = {"upload_url": "https://signed.url"}
        mock_async_create_signed_upload_url.return_value = mock_signed_url_response

        # Set up mock API client
        mock_api_client = MagicMock()

        # Check that the response matches the expected response
        with pytest.raises(DarwinException):
            asyncio.run(
                uploads.async_register_and_create_signed_upload_url(
                    mock_api_client,
                    "my-team",
                    "my-dataset",
                    [(Mock(), Mock())],
                    False,
                    False,
                    False,
                )
            )


class TestConfirmUpload(SetupTests):
    @responses.activate
    def test_async_confirm_upload(
        self, base_client: ClientCore, default_url: str
    ) -> None:
        # Call the function with mocked arguments
        responses.add(
            "POST",
            f"{default_url}/uploads/123/confirm",
            status=200,
            json={},
        )

        actual_response = asyncio.run(
            uploads.async_confirm_upload(
                base_client,
                "my-team",
                "123",
            )
        )

        assert actual_response is None  # Function doesn't return anything on success

    @responses.activate
    def test_async_confirm_upload_raises_on_returned_errors(
        self, base_client: ClientCore, default_url: str
    ) -> None:
        # Call the function with mocked arguments
        responses.add(
            "POST",
            f"{default_url}/uploads/123/confirm",
            status=200,  # API will not normally return this on success, but we're not test API
            json={"errors": ["error1", "error2"]},
        )

        with pytest.raises(DarwinException) as exc:
            asyncio.run(
                uploads.async_confirm_upload(
                    base_client,
                    "my-team",
                    "123",
                )
            )

        assert (
            "Failed to confirm upload in darwin.future.core.items.uploads: ['error1', 'error2']"
            in str(exc)
        )

    def test_async_confirm_upload_raises(self, base_client: ClientCore) -> None:
        base_client.post = MagicMock()  # type: ignore
        base_client.post.side_effect = DarwinException("Error")

        with pytest.raises(DarwinException):
            asyncio.run(uploads.async_confirm_upload(base_client, "team", "123"))


class TestSynchronousMethods:
    @pytest.fixture
    def mock_async_register_upload(self) -> Generator:
        with patch.object(uploads, "async_register_upload") as mock:
            yield mock

    @pytest.fixture
    def mock_async_create_signed_upload_url(self) -> Generator:
        with patch.object(uploads, "async_create_signed_upload_url") as mock:
            yield mock

    @pytest.fixture
    def mock_async_register_and_create_signed_upload_url(self) -> Generator:
        with patch.object(
            uploads, "async_register_and_create_signed_upload_url"
        ) as mock:
            yield mock

    @pytest.fixture
    def mock_async_confirm_upload(self) -> Generator:
        with patch.object(uploads, "async_confirm_upload") as mock:
            yield mock

    def test_register_upload(
        self,
        mock_async_register_upload: MagicMock,
        base_client: ClientCore,
    ) -> None:
        uploads.register_upload(base_client, "team", "dataset", [(Mock(), Mock())])

        mock_async_register_upload.assert_called_once()

    def test_create_signed_upload_url(
        self,
        mock_async_create_signed_upload_url: MagicMock,
        base_client: ClientCore,
    ) -> None:
        uploads.create_signed_upload_url(base_client, "team", "123")

        mock_async_create_signed_upload_url.assert_called_once()

    def test_register_and_create_signed_upload_url(
        self,
        mock_async_register_and_create_signed_upload_url: MagicMock,
        base_client: ClientCore,
    ) -> None:
        uploads.register_and_create_signed_upload_url(
            base_client, "team", "dataset", [(Mock(), Mock())]
        )

        mock_async_register_and_create_signed_upload_url.assert_called_once()

    def test_confirm_upload(
        self,
        mock_async_confirm_upload: MagicMock,
        base_client: ClientCore,
    ) -> None:
        uploads.confirm_upload(base_client, "team", "123")

        mock_async_confirm_upload.assert_called_once()


if __name__ == "__main__":
    pytest.main(["-s", "-v", __file__])



from typing import List, Optional, Tuple

from pydantic import ValidationError

from darwin.future.core.client import ClientCore
from darwin.future.data_objects.team import TeamCore, TeamMemberCore


def get_team(client: ClientCore, team_slug: Optional[str] = None) -> TeamCore:
    """
    Returns a TeamCore object for the specified team slug.

    Parameters:
        client (ClientCore): The client to use for the request.
        team_slug (Optional[str]): The slug of the team to get. If not specified, the
            default team from the client's config will be used.

    Returns:
        TeamCore: The TeamCore object for the specified team slug.

    Raises:
        HTTPError: If the response status code is not in the 200-299 range.
    """
    if not team_slug:
        team_slug = client.config.default_team
    response = client.get(f"/teams/{team_slug}/")
    return TeamCore.model_validate(response)


def get_team_members(
    client: ClientCore,
) -> Tuple[List[TeamMemberCore], List[ValidationError]]:
    """
    Returns a tuple containing a list of TeamMemberCore objects and a list of exceptions
    that occurred while parsing the response.

    Parameters:
        client (ClientCore): The client to use for the request.

    Returns:
        List[TeamMemberCore]:
            List of TeamMembers
        List[ValidationError]:
            List of ValidationError on failed objects

    Raises:
        HTTPError: If the response status code is not in the 200-299 range.
    """
    response = client.get("/memberships")
    members = []
    errors = []
    for item in response:
        try:
            members.append(TeamMemberCore.model_validate(item))
        except ValidationError as e:
            errors.append(e)
    return members, errors



import string

from pytest import mark, raises
from requests import HTTPError
from responses import RequestsMock

from darwin.future.core.client import DarwinConfig
from darwin.future.meta.client import Client
from darwin.future.meta.objects.dataset import Dataset
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.objects.fixtures import *

# `datasets` tests
# TODO datasets tests

# `get_dataset_by_id` tests
# TODO get_dataset_by_id tests


# `create_dataset` tests
def test_create_dataset_raises_HTTPError(base_config: DarwinConfig) -> None:
    valid_client = Client(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps, raises(HTTPError):
        rsps.add(rsps.POST, base_url, status=500)
        Dataset.create_dataset(valid_client, valid_slug)


def test_create_dataset_returns_dataset_created_if_dataset_created(
    base_config: DarwinConfig,
) -> None:
    valid_client = Client(base_config)
    valid_slug = "test_dataset"

    base_url = base_config.base_url + "api/datasets"

    with RequestsMock() as rsps:
        rsps.add(
            rsps.POST,
            base_url,
            json={"id": 1, "name": "Test Dataset", "slug": "test_dataset"},
            status=201,
        )

        dataset_created = Dataset.create_dataset(valid_client, valid_slug)

        assert dataset_created is not None
        assert dataset_created.id == 1
        assert dataset_created.name == "test dataset"
        assert dataset_created.slug == "test_dataset"


# `update_dataset` tests
# TODO update_dataset tests


@mark.parametrize(
    "invalid_slug",
    [
        "",
        " ",
        "test dataset",
        *[f"dataset_{c}" for c in string.punctuation if c not in ["-", "_", "."]],
    ],
)
def test_validate_slugh_raises_exception_if_passed_invalid_inputs(
    invalid_slug: str,
) -> None:
    with raises(AssertionError):
        Dataset._validate_slug(invalid_slug)


def test_validate_slug_returns_none_if_passed_valid_slug() -> None:
    valid_slug = "test-dataset"

    assert Dataset._validate_slug(valid_slug) is None


def test_delete(base_meta_dataset: Dataset, base_config: DarwinConfig) -> None:
    base_url = base_config.base_url + "api/datasets"
    with RequestsMock() as rsps:
        rsps.add(
            rsps.PUT,
            base_url + f"/{base_meta_dataset.id}/archive",
            json={
                "id": base_meta_dataset.id,
                "name": "Test Dataset",
                "slug": "test_dataset",
            },
            status=200,
        )
        dataset_deleted = base_meta_dataset.delete()

        assert dataset_deleted == 1


def test_dataset_str_method(base_meta_dataset: Dataset) -> None:
    assert (
        base_meta_dataset.__str__()
        == "Dataset\n\
- Name: test dataset\n\
- Dataset Slug: test-dataset\n\
- Dataset ID: 1\n\
- Dataset Releases: No releases"
    )


def test_dataset_repr_method(base_meta_dataset: Dataset) -> None:
    assert base_meta_dataset.__repr__() == str(base_meta_dataset)



from typing import Optional, Union

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONDict
from darwin.future.data_objects.properties import FullProperty


def create_property(
    client: ClientCore,
    params: Union[FullProperty, JSONDict],
    team_slug: Optional[str] = None,
) -> FullProperty:
    """
    Creates a property for the specified team slug.

    Parameters:
        client (ClientCore): The client to use for the request.
        team_slug (Optional[str]): The slug of the team to get. If not specified, the
            default team from the client's config will be used.
        params (Optional[JSONType]): The JSON data to use for the request.

    Returns:
        FullProperty: FullProperty object for the created property.

    Raises:
        HTTPError: If the response status code is not in the 200-299 range.
    """
    if not team_slug:
        team_slug = client.config.default_team
    if isinstance(params, FullProperty):
        params = params.to_create_endpoint()
    response = client.post(f"/v2/teams/{team_slug}/properties", data=params)
    assert isinstance(response, dict)
    return FullProperty.model_validate(response)



import json

import pytest

from darwin.future.data_objects.darwinV2 import (
    BoundingBoxAnnotation,
    DarwinV2,
    EllipseAnnotation,
    PolygonAnnotation,
)


@pytest.fixture
def raw_json() -> dict:
    with open("./darwin/future/tests/data/base_annotation.json") as f:
        raw_json = json.load(f)
    return raw_json


def test_loads_base_darwin_v2(raw_json: dict) -> None:
    test = DarwinV2.model_validate(raw_json)
    assert len(test.annotations) == 3
    assert isinstance(test.annotations[0], BoundingBoxAnnotation)
    assert isinstance(test.annotations[1], EllipseAnnotation)
    assert isinstance(test.annotations[2], PolygonAnnotation)


def test_bbox_annotation(raw_json: dict) -> None:
    bounds_annotation = raw_json["annotations"][0]
    BoundingBoxAnnotation.model_validate(bounds_annotation)


def test_ellipse_annotation(raw_json: dict) -> None:
    ellipse_annotation = raw_json["annotations"][1]
    EllipseAnnotation.model_validate(ellipse_annotation)


def test_polygon_annotation(raw_json: dict) -> None:
    polygon_annotation = raw_json["annotations"][2]
    PolygonAnnotation.model_validate(polygon_annotation)


def test_polygon_bbx_validator(raw_json: dict) -> None:
    polygon_annotation = raw_json["annotations"][2]
    without_bbx = polygon_annotation.copy()
    del without_bbx["bounding_box"]
    without_bb_annotation = PolygonAnnotation.model_validate(without_bbx)
    with_bb_annotation = PolygonAnnotation.model_validate(polygon_annotation)

    assert without_bb_annotation.bounding_box is not None
    assert with_bb_annotation.bounding_box is not None
    assert without_bb_annotation == with_bb_annotation
    bounds_annotation = raw_json["annotations"][0]
    BoundingBoxAnnotation.model_validate(bounds_annotation)



from __future__ import annotations

from typing import List, Literal, Tuple, Union
from uuid import UUID

from pydantic import ValidationError

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONDict, QueryString
from darwin.future.data_objects.item import Folder, ItemCore


def get_item_ids(
    api_client: ClientCore,
    team_slug: str,
    dataset_id: Union[str, int],
    params: QueryString = QueryString({}),
) -> List[UUID]:
    """
    Returns a list of item ids for the dataset

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get item ids for
    dataset_id: str
        The id or slug of the dataset to get item ids for

    Returns
    -------
    List[UUID]
        A list of item ids
    """
    response = api_client.get(
        f"/v2/teams/{team_slug}/items/list_ids",
        QueryString(
            {
                "dataset_ids": str(dataset_id),
            }
        )
        + params,
    )
    assert isinstance(response, dict)
    uuids = [UUID(uuid) for uuid in response["item_ids"]]
    return uuids


def get_item_ids_stage(
    api_client: ClientCore,
    team_slug: str,
    dataset_id: Union[int, str],
    stage_id: Union[UUID, str],
) -> List[UUID]:
    """
    Returns a list of item ids for the stage

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get item ids for
    dataset_id: str
        The id or slug of the dataset to get item ids for
    stage_id: str
        The id or slug of the stage to get item ids for

    Returns
    -------
    List[UUID]
        A list of item ids
    """
    response = api_client.get(
        f"/v2/teams/{team_slug}/items/ids",
        QueryString(
            {"workflow_stage_ids": str(stage_id), "dataset_ids": str(dataset_id)}
        ),
    )
    assert isinstance(response, dict)
    uuids = [UUID(uuid) for uuid in response["item_ids"]]
    return uuids


def get_item(
    api_client: ClientCore,
    team_slug: str,
    item_id: Union[UUID, str],
    params: QueryString = QueryString({}),
) -> ItemCore:
    """
    Returns an item

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get item ids for
    item_id: str
        The id or slug of the item to get

    Returns
    -------
    Item
        An item object
    """
    response = api_client.get(f"/v2/teams/{team_slug}/items/{item_id}", params)
    assert isinstance(response, dict)
    return ItemCore.model_validate(response)


def list_items(
    api_client: ClientCore,
    team_slug: str,
    dataset_ids: int | list[int] | Literal["all"],
    params: QueryString = QueryString({}),
) -> Tuple[List[ItemCore], List[ValidationError]]:
    """
    Returns a list of items for the dataset

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get items for
    dataset_id: str
        The id or slug of the dataset to get items for

    Returns
    -------
    List[Item]
        A list of items
    List[ValidationError]
        A list of ValidationError on failed objects
    """
    dataset_ids = (
        dataset_ids
        if isinstance(dataset_ids, list) or dataset_ids == "all"
        else [dataset_ids]
    )
    params = params + QueryString({"dataset_ids": dataset_ids})
    response = api_client.get(f"/v2/teams/{team_slug}/items", params)
    assert isinstance(response, dict)
    items: List[ItemCore] = []
    exceptions: List[ValidationError] = []
    for item in response["items"]:
        assert isinstance(item, dict)
        try:
            items.append(ItemCore.model_validate(item))
        except ValidationError as e:
            exceptions.append(e)
    return items, exceptions


def list_folders(
    api_client: ClientCore,
    team_slug: str,
    params: QueryString,
) -> Tuple[List[Folder], List[ValidationError]]:
    """
    Returns a list of folders for the team and dataset

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get folder ids for
    params: QueryString
        parameters to filter the folders

    Returns
    -------
    List[Folder]
        The folders
    List[ValidationError]
        A list of ValidationError on failed objects
    """
    assert "dataset_ids" in params.value, "dataset_ids must be provided"
    response = api_client.get(f"/v2/teams/{team_slug}/items/folders", params)
    assert isinstance(response, dict)
    assert "folders" in response
    exceptions: List[ValidationError] = []
    folders: List[Folder] = []
    for item in response["folders"]:
        try:
            folders.append(Folder.model_validate(item))
        except ValidationError as e:
            exceptions.append(e)
    return folders, exceptions


def list_items_unstable(
    api_client: ClientCore,
    team_slug: str,
    params: JSONDict,
) -> Tuple[List[ItemCore], List[ValidationError]]:
    """
    Returns a list of items for the dataset from the advanced filters 'unstable' endpoint

    Parameters
    ----------
    client: Client
        The client to use for the request
    team_slug: str
        The slug of the team to get items for
    params: JSONType
        Must include at least dataset_ids

    Returns
    -------
    List[Item]
        A list of items
    List[ValidationError]
        A list of ValidationError on failed objects
    """
    if "dataset_ids" not in params:
        raise ValueError("dataset_ids must be provided")
    response = api_client.post(f"/unstable/teams/{team_slug}/items/list", params)
    assert isinstance(response, dict)
    items: List[ItemCore] = []
    exceptions: List[ValidationError] = []
    for item in response["items"]:
        assert isinstance(item, dict)
        try:
            items.append(ItemCore.model_validate(item))
        except ValidationError as e:
            exceptions.append(e)
    return items, exceptions



import responses

from darwin.future.core.client import DarwinConfig
from darwin.future.data_objects.team import TeamCore
from darwin.future.meta.client import Client
from darwin.future.meta.objects.team import Team
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.meta.fixtures import *


def test_creates_from_api_key() -> None:
    with responses.RequestsMock() as rsps:
        base_api_endpoint = DarwinConfig._default_api_endpoint()
        rsps.add(
            responses.GET,
            base_api_endpoint + "users/token_info",
            json={"selected_team": {"slug": "test-team"}},
        )
        client = Client.from_api_key(api_key="test")
        assert client.config.default_team == "test-team"


def test_team_property(
    base_meta_client: Client, base_team: TeamCore, base_team_json: dict
) -> None:
    client = base_meta_client
    endpoint = client.config.api_endpoint + f"teams/{client.config.default_team}"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json=base_team_json)
        team = client.team
        assert isinstance(team, Team)
        assert team._element == base_team



from typing import Dict

import pytest
import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.items.untag_items import untag_items
from darwin.future.data_objects.typing import UnknownType
from darwin.future.exceptions import BadRequest
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_untag_items(base_client: ClientCore) -> None:
    team_slug = "test-team"
    dataset_ids = [1, 2, 3]
    tag_id = 123456
    item_ids = [
        "00000000-0000-0000-0000-000000000001",
        "00000000-0000-0000-0000-000000000002",
    ]
    filters = {"item_ids": item_ids}

    responses.add(
        responses.DELETE,
        base_client.config.api_endpoint + "v2/teams/test-team/items/slots/tags",
        json={"created_commands": 2},
        status=200,
    )

    response = untag_items(
        client=base_client,
        team_slug=team_slug,
        dataset_ids=dataset_ids,
        tag_id=tag_id,
        filters=filters,
    )

    assert response == {"created_commands": 2}


@responses.activate
def test_untag_items_empty_filters_error(base_client: ClientCore) -> None:
    team_slug = "test-team"
    dataset_ids = [1, 2, 3]
    tag_id = 123456
    # this should raise an error as no filters are provided
    filters: Dict[str, UnknownType] = {}

    responses.add(
        responses.DELETE,
        base_client.config.api_endpoint + "v2/teams/test-team/items/slots/tags",
        json={"created_commands": 2},
        status=200,
    )

    with pytest.raises(AssertionError) as excinfo:
        untag_items(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            tag_id=tag_id,
            filters=filters,
        )
        (msg,) = excinfo.value.args
        assert (
            msg
            == "No parameters provided, please provide at least one non-dataset id filter"
        )


@responses.activate
def test_untag_items_bad_request_error(base_client: ClientCore) -> None:
    team_slug = "test-team"
    dataset_ids = [1, 2, 3]
    tag_id = "123456"
    item_ids = [
        "00000000-0000-0000-0000-000000000001",
        "00000000-0000-0000-0000-000000000002",
    ]
    filters = {"item_ids": item_ids}

    responses.add(
        responses.DELETE,
        base_client.config.api_endpoint + "v2/teams/test-team/items/slots/tags",
        json={"error": "Bad Request"},
        status=400,
    )

    with pytest.raises(BadRequest):
        untag_items(
            client=base_client,
            team_slug=team_slug,
            dataset_ids=dataset_ids,
            tag_id=tag_id,  # type: ignore
            filters=filters,
        )



from typing import List
from uuid import UUID

import responses
from pytest import fixture
from responses.matchers import query_param_matcher

from darwin.future.data_objects.workflow import WFEdgeCore, WFStageCore, WFTypeCore
from darwin.future.meta.client import Client
from darwin.future.meta.objects.stage import Stage
from darwin.future.meta.queries.item import ItemQuery
from darwin.future.meta.queries.item_id import ItemIDQuery
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.core.items.fixtures import *
from darwin.future.tests.meta.fixtures import *


@fixture
def uuid_str() -> str:
    return "00000000-0000-0000-0000-000000000000"


@fixture
def base_WFStage(uuid_str: str) -> WFStageCore:
    return WFStageCore(
        id=UUID(uuid_str),
        name="test-stage",
        type=WFTypeCore.ANNOTATE,
        assignable_users=[],
        edges=[],
    )


@fixture
def stage_meta(
    base_meta_client: Client, base_WFStage: WFStageCore, workflow_id: UUID
) -> Stage:
    return Stage(
        client=base_meta_client,
        element=base_WFStage,
        meta_params={
            "team_slug": "default-team",
            "dataset_id": 1337,
            "workflow_id": workflow_id,
        },
    )


def test_item_ids(
    base_meta_client: Client, stage_meta: Stage, UUIDs_str: List[str], UUIDs: List[UUID]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_meta_client.config.api_endpoint
            + "v2/teams/default-team/items/list_ids",
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "500",
                        "workflow_stage_ids": str(stage_meta.id),
                        "dataset_ids": "1337",
                    }
                )
            ],
            json={"item_ids": UUIDs_str},
            status=200,
        )
        item_ids = [x.id for x in stage_meta.item_ids.collect_all()]
        assert item_ids == UUIDs


def test_move_attached_files_to_stage(
    base_meta_client: Client, stage_meta: Stage, UUIDs_str: List[str], UUIDs: List[UUID]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_meta_client.config.api_endpoint
            + "v2/teams/default-team/items/list_ids",
            json={"item_ids": UUIDs_str},
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "500",
                        "workflow_stage_ids": str(stage_meta.id),
                        "dataset_ids": "1337",
                    }
                )
            ],
            status=200,
        )
        rsps.add(
            rsps.POST,
            base_meta_client.config.api_endpoint + "v2/teams/default-team/items/stage",
            json={"success": UUIDs_str},
            status=200,
        )
        stage_meta.move_attached_files_to_stage(stage_meta.id, wait=False)


def test_move_attached_files_to_stage_wait(
    base_meta_client: Client, stage_meta: Stage, UUIDs_str: List[str], UUIDs: List[UUID]
) -> None:
    with responses.RequestsMock() as rsps:
        rsps.add(
            rsps.GET,
            base_meta_client.config.api_endpoint
            + "v2/teams/default-team/items/list_ids",
            json={"item_ids": UUIDs_str},
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "500",
                        "workflow_stage_ids": str(stage_meta.id),
                        "dataset_ids": "1337",
                    }
                )
            ],
            status=200,
        )
        rsps.add(
            rsps.POST,
            base_meta_client.config.api_endpoint + "v2/teams/default-team/items/stage",
            json={"success": UUIDs_str},
            status=200,
        )
        for uuid in stage_meta.item_ids.collect_all():
            rsps.add(
                rsps.GET,
                base_meta_client.config.api_endpoint
                + f"v2/teams/default-team/items/{uuid}",
                json={
                    "archived": False,
                    "dataset_id": 1337,
                    "id": "00000000-0000-0000-0000-000000000000",
                    "layout": None,
                    "name": "test_0",
                    "path": "test_path",
                    "priority": 0,
                    "processing_status": "complete",
                    "slots": [],
                    "tags": [],
                },
                status=200,
            )
        stage_meta.move_attached_files_to_stage(
            stage_meta.id, wait=True, wait_max_attempts=5, wait_time=0.5
        )


def test_get_stage_id(stage_meta: Stage) -> None:
    assert stage_meta.id == UUID("00000000-0000-0000-0000-000000000000")


def test_get_stage_name(stage_meta: Stage) -> None:
    assert stage_meta.name == "test-stage"


def test_get_stage_type(stage_meta: Stage) -> None:
    assert stage_meta.type == "annotate"


def test_get_stage_edges(stage_meta: Stage) -> None:
    edges = [
        WFEdgeCore(
            name="edge_1",
            id=UUID("00000000-0000-0000-0000-000000000000"),
            source_stage_id=UUID("00000000-0000-0000-0000-000000000000"),
            target_stage_id=UUID("00000000-0000-0000-0000-000000000000"),
        ),
        WFEdgeCore(
            name="edge_2",
            id=UUID("00000000-0000-0000-0000-000000000000"),
            source_stage_id=UUID("00000000-0000-0000-0000-000000000000"),
            target_stage_id=UUID("00000000-0000-0000-0000-000000000000"),
        ),
    ]
    test_stage = Stage(
        client=stage_meta.client,
        element=WFStageCore(
            id=UUID("00000000-0000-0000-0000-000000000000"),
            name="test-stage",
            type=WFTypeCore.ANNOTATE,
            assignable_users=[],
            edges=edges,
        ),
        meta_params={
            "team_slug": "default-team",
            "dataset_id": 000000,
            "workflow_id": UUID("00000000-0000-0000-0000-000000000000"),
        },
    )
    assert len(test_stage.edges) == 2
    assert test_stage.edges[0].name == "edge_1"
    assert test_stage.edges[0].id == UUID("00000000-0000-0000-0000-000000000000")
    assert test_stage.edges[0].source_stage_id == UUID(
        "00000000-0000-0000-0000-000000000000"
    )
    assert test_stage.edges[0].target_stage_id == UUID(
        "00000000-0000-0000-0000-000000000000"
    )


def test_stage_str_method(stage_meta: Stage) -> None:
    assert (
        str(stage_meta)
        == "Stage\n\
- Stage Name: test-stage\n\
- Stage Type: annotate\n\
- Stage ID: 00000000-0000-0000-0000-000000000000"
    )


def test_stage_repr_method(stage_meta: Stage) -> None:
    assert repr(stage_meta) == str(stage_meta)


def test_has_item_properties(stage_meta: Stage) -> None:
    assert isinstance(stage_meta.items, ItemQuery)
    assert isinstance(stage_meta.item_ids, ItemIDQuery)
    assert isinstance(stage_meta.id, UUID)
    assert isinstance(stage_meta.name, str)
    assert isinstance(stage_meta.type, str)
    assert isinstance(stage_meta.edges, list)



from __future__ import annotations

from typing import Dict, List

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONType
from darwin.future.data_objects.typing import UnknownType


def move_list_of_items_to_folder(
    client: ClientCore,
    team_slug: str,
    dataset_ids: int | List[int],
    path: str,
    filters: Dict[str, UnknownType] = {},
) -> JSONType:
    """
    Move specified items to a folder

    Parameters
    ----------
    client: Client
        The client to use for the request.
    team_slug: str
        The slug of the team containing the items.
    dataset_ids: int | List[int]
        The ID(s) of the dataset(s) containing the items.
    path: str
        The path to the folder to move the items to.
    filters: Dict[str, UnknownType]
        Filter parameters.

    Returns
    -------
    JSONType
        The response data.
    """
    assert (
        filters
    ), "No parameters provided, please provide at least one non-dataset id filter"
    payload = {
        "filters": {
            "dataset_ids": (
                dataset_ids if isinstance(dataset_ids, list) else [dataset_ids]
            ),
            **filters,
        },
        "path": path,
    }

    return client.post(f"/v2/teams/{team_slug}/items/path", data=payload)



from functools import reduce
from typing import Dict

from darwin.future.core.items.get import get_item_ids
from darwin.future.core.types.common import QueryString
from darwin.future.core.types.query import PaginatedQuery, QueryFilter
from darwin.future.data_objects.sorting import SortingMethods
from darwin.future.meta.objects.v7_id import V7ID


class ItemIDQuery(PaginatedQuery[V7ID]):
    def _collect(self) -> Dict[int, V7ID]:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query item ids")
        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_id to query item ids")
        team_slug: str = self.meta_params["team_slug"]
        dataset_ids: int = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        params: QueryString = reduce(
            lambda s1, s2: s1 + s2,
            [
                self.page.to_query_string(),
                *[QueryString(f.to_dict()) for f in self.filters],
            ],
        )
        uuids = get_item_ids(self.client, team_slug, dataset_ids, params)

        results = {
            i
            + self.page.offset: V7ID(
                client=self.client, element=uuid, meta_params=self.meta_params
            )
            for i, uuid in enumerate(uuids)
        }
        return results

    def sort(self, **kwargs: str) -> "ItemIDQuery":
        valid_values = {"asc", "desc"}
        for value in kwargs.values():
            if value not in valid_values:
                raise ValueError(
                    f"Invalid sort value: {value}. Must be one of {valid_values}."
                )
        sorting_methods = SortingMethods(**kwargs)  # type: ignore
        for key, value in sorting_methods.dict().items():
            if value is not None:
                filter = QueryFilter(name=f"sort[{key}]", param=value)
                self.filters.append(filter)
        return self



from __future__ import annotations

from functools import reduce
from typing import Dict, Literal, Protocol, Union

from darwin.future.core.items.archive_items import archive_list_of_items
from darwin.future.core.items.assign_items import assign_items
from darwin.future.core.items.delete_items import delete_list_of_items
from darwin.future.core.items.get import list_items, list_items_unstable
from darwin.future.core.items.move_items_to_folder import move_list_of_items_to_folder
from darwin.future.core.items.restore_items import restore_list_of_items
from darwin.future.core.items.set_item_layout import set_item_layout
from darwin.future.core.items.set_item_priority import set_item_priority
from darwin.future.core.items.set_stage_to_items import set_stage_to_items
from darwin.future.core.items.tag_items import tag_items
from darwin.future.core.items.untag_items import untag_items
from darwin.future.core.types.common import JSONDict, QueryString
from darwin.future.core.types.query import PaginatedQuery, QueryFilter
from darwin.future.data_objects.advanced_filters import GroupFilter, SubjectFilter
from darwin.future.data_objects.item import ItemLayout
from darwin.future.data_objects.sorting import SortingMethods
from darwin.future.data_objects.workflow import WFStageCore
from darwin.future.exceptions import BadRequest
from darwin.future.meta.objects.item import Item


class hasStage(Protocol):
    # Using Protocol to avoid circular imports between item.py and stage.py
    _element: WFStageCore


class ItemQuery(PaginatedQuery[Item]):
    def _collect(self) -> Dict[int, Item]:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        params = self._build_params()
        if isinstance(params, QueryString):
            items_core, errors = list_items(self.client, team_slug, dataset_ids, params)
        else:
            items_core, errors = list_items_unstable(
                api_client=self.client, team_slug=team_slug, params=params
            )
        offset = self.page.offset
        items = {
            i
            + offset: Item(
                client=self.client, element=item, meta_params=self.meta_params
            )
            for i, item in enumerate(items_core)
        }
        return items

    def sort(self, **kwargs: str) -> ItemQuery:
        valid_values = {"asc", "desc"}
        for value in kwargs.values():
            if value not in valid_values:
                raise ValueError(
                    f"Invalid sort value: {value}. Must be one of {valid_values}."
                )
        sorting_methods = SortingMethods(**kwargs)  # type: ignore
        for key, value in sorting_methods.dict().items():
            if value is not None:
                filter = QueryFilter(name=f"sort[{key}]", param=value)
                self.filters.append(filter)
        return self

    def delete(self) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        delete_list_of_items(self.client, team_slug, dataset_ids, filters)

    def move_to_folder(self, path: str) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")

        if not path:
            raise ValueError("Must specify path to move items to")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        move_list_of_items_to_folder(self.client, team_slug, dataset_ids, path, filters)

    def _build_params(self) -> Union[QueryString, JSONDict]:
        if self._advanced_filters is None:
            return reduce(
                lambda s1, s2: s1 + s2,
                [
                    self.page.to_query_string(),
                    *[QueryString(f.to_dict()) for f in self.filters],
                ],
            )
        if not self.meta_params["dataset_ids"] and not self.meta_params["dataset_id"]:
            raise ValueError("Must specify dataset_ids to query items")
        dataset_id = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        return {
            "dataset_ids": [dataset_id],
            "page": self.page.model_dump(),
            "filter": self._advanced_filters.model_dump(),
        }

    def set_priority(self, priority: int) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")

        if not priority:
            raise ValueError("Must specify priority to set items to")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        set_item_priority(self.client, team_slug, dataset_ids, priority, filters)

    def restore(self) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        restore_list_of_items(self.client, team_slug, dataset_ids, filters)

    def archive(self) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        archive_list_of_items(self.client, team_slug, dataset_ids, filters)

    def set_layout(self, layout: ItemLayout) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")

        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")

        assert isinstance(layout, ItemLayout)
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        set_item_layout(self.client, team_slug, dataset_ids, layout, filters)

    def tag(self, tag_id: int) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")
        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        if not tag_id:
            raise ValueError("Must specify tag_id to tag items with")
        if not isinstance(tag_id, int):
            raise BadRequest(f"tag_id must be an integer, got {type(tag_id)}")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        tag_items(self.client, team_slug, dataset_ids, tag_id, filters)

    def untag(self, tag_id: int) -> None:
        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")
        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        if not tag_id:
            raise ValueError("Must specify tag_id to untag items with")
        if not isinstance(tag_id, int):
            raise BadRequest(f"tag_id must be an integer, got {type(tag_id)}")
        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}
        untag_items(self.client, team_slug, dataset_ids, tag_id, filters)

    def assign(self, assignee_id: int, workflow_id: str | None = None) -> None:
        if not assignee_id:
            raise ValueError("Must specify assignee to assign items to")

        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")
        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        if not workflow_id:
            # if workflow_id is not specified, get it from the meta_params
            # this will be present in the case of a workflow object
            if "workflow_id" in self.meta_params:
                workflow_id = str(self.meta_params["workflow_id"])
            else:
                raise ValueError("Must specify workflow_id to set items to")
        assert isinstance(workflow_id, str)

        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}

        assign_items(
            self.client, team_slug, dataset_ids, assignee_id, workflow_id, filters
        )

    def set_stage(
        self, stage_or_stage_id: hasStage | str, workflow_id: str | None = None
    ) -> None:
        if not stage_or_stage_id:
            raise ValueError(
                "Must specify stage (either Stage object or stage_id string) to set items to"
            )

        if "team_slug" not in self.meta_params:
            raise ValueError("Must specify team_slug to query items")
        if (
            "dataset_ids" not in self.meta_params
            and "dataset_id" not in self.meta_params
        ):
            raise ValueError("Must specify dataset_ids to query items")
        if not workflow_id:
            # if workflow_id is not specified, get it from the meta_params
            # this will be present in the case of a workflow object
            if "workflow_id" in self.meta_params:
                workflow_id = str(self.meta_params["workflow_id"])
            else:
                raise ValueError("Must specify workflow_id to set items to")
        assert isinstance(workflow_id, str)
        if not stage_or_stage_id:
            raise ValueError("Must specify stage to set stage for items")

        # get stage_id from stage_or_stage_id
        if isinstance(stage_or_stage_id, str):
            stage_id = stage_or_stage_id
        else:
            stage_id = str(stage_or_stage_id._element.id)

        dataset_ids = (
            self.meta_params["dataset_ids"]
            if "dataset_ids" in self.meta_params
            else self.meta_params["dataset_id"]
        )
        team_slug = self.meta_params["team_slug"]
        self.collect_all()
        ids = [item.id for item in self]
        filters = {"item_ids": [str(item) for item in ids]}

        set_stage_to_items(
            self.client, team_slug, dataset_ids, stage_id, workflow_id, filters
        )

    def where(
        self,
        *args: GroupFilter | SubjectFilter,
        _operator: Literal["and", "or"] = "and",
        **kwargs: str,
    ) -> ItemQuery:
        """Adds a filter to the query
        This can be used with simple filters via a keyword argument
        or with advanced filters via a GroupFilter or SubjectFilter object as args.

        Args:
            _operator (Literal["and", "or"], optional): The operator to use when combining

        Raises:
            ValueError: Raises if trying to use both simple and advanced filters

        Returns:
            ItemQuery: Self
        """
        if len(args) > 0 and len(kwargs) > 0:
            raise ValueError("Cannot specify both args and kwargs")
        if len(kwargs) > 1:
            return super().where(**kwargs)
        arg_list = list(args)
        if self._advanced_filters is None:
            self._advanced_filters = arg_list.pop(0)
        for arg in arg_list:
            if _operator == "and":
                self._advanced_filters = self._advanced_filters & arg
            if _operator == "or":
                self._advanced_filters = self._advanced_filters | arg
        return self



import sys
import uuid
import warnings
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Any

from rich.console import Console

from darwin.utils import attempt_decode

console = Console()
try:
    import cc3d
    import nibabel as nib
    from scipy.ndimage import zoom
except ImportError:
    import_fail_string = r"""
    You must install darwin-py with pip install darwin-py\[medical]
    in order to import using nifti format
    """
    console.print(import_fail_string)
    sys.exit(1)
import numpy as np
from jsonschema import validate
from upolygon import find_contours

import darwin.datatypes as dt
from darwin.importer.formats.nifti_schemas import nifti_import_schema


def parse_path(
    path: Path,
    legacy_remote_file_slot_affine_maps: Dict[Path, Dict[str, Any]] = {},
    pixdims_and_primary_planes: Dict[Path, Dict[str, Tuple[List[float], str]]] = {},
) -> Optional[List[dt.AnnotationFile]]:
    """
    Parses the given ``nifti`` file and returns a ``List[dt.AnnotationFile]`` with the parsed
    information.

    Parameters
    ----------
    path : Path
        The ``Path`` to the ``nifti`` file.
    legacy_remote_file_slot_affine_maps : Optional[Dict[Path, Dict[str, Any]]]
        A dictionary of remote file full paths to their slot affine maps
    remote_file_pixdims : Optional[Dict[Path, Dict[str, Tuple[List[float], str]]]]
        A dictionary of remote file full paths to their pixel dimensions and primary plane

    Returns
    -------
    Optional[List[dt.AnnotationFile]]
        Returns ``None`` if the given file is not in ``json`` format, or ``List[dt.AnnotationFile]``
        otherwise.
    """
    if not isinstance(path, Path):
        path = Path(path)
    if path.suffix != ".json":
        console.print(
            "Skipping file: {} (not a json file)".format(path), style="bold yellow"
        )
        return None
    data = attempt_decode(path)
    try:
        validate(data, schema=nifti_import_schema)
    except Exception:
        console.print(
            "Skipping file: {} (invalid json file, see schema for details)".format(
                path
            ),
            style="bold yellow",
        )
        return None
    nifti_annotations = data.get("data")
    if nifti_annotations is None or nifti_annotations == []:
        console.print(
            "Skipping file: {} (no data found)".format(path), style="bold yellow"
        )
        return None
    annotation_files = []
    for nifti_annotation in nifti_annotations:
        remote_file_path = Path(nifti_annotation["image"])
        if not str(remote_file_path).startswith("/"):
            remote_file_path = Path("/" + str(remote_file_path))
        annotation_file = _parse_nifti(
            Path(nifti_annotation["label"]),
            Path(nifti_annotation["image"]),
            path,
            class_map=nifti_annotation.get("class_map"),
            mode=nifti_annotation.get("mode", "image"),
            slot_names=nifti_annotation.get("slot_names", []),
            is_mpr=nifti_annotation.get("is_mpr", False),
            remote_file_path=remote_file_path,
            legacy_remote_file_slot_affine_maps=legacy_remote_file_slot_affine_maps,
            pixdims_and_primary_planes=pixdims_and_primary_planes,
        )
        annotation_files.append(annotation_file)
    return annotation_files


def _parse_nifti(
    nifti_path: Path,
    filename: Path,
    json_path: Path,
    class_map: Dict,
    mode: str,
    slot_names: List[str],
    is_mpr: bool,
    remote_file_path: Path,
    legacy_remote_file_slot_affine_maps: Dict[Path, Dict[str, Any]] = {},
    pixdims_and_primary_planes: Dict[Path, Dict[str, Tuple[List[float], str]]] = {},
) -> dt.AnnotationFile:
    img = process_nifti(
        nib.load(nifti_path),
        remote_file_path=remote_file_path,
        legacy_remote_file_slot_affine_maps=legacy_remote_file_slot_affine_maps,
    )
    legacy = remote_file_path in legacy_remote_file_slot_affine_maps
    processed_class_map = process_class_map(class_map)
    video_annotations = []
    if mode == "instances":  # For each instance produce a video annotation
        for class_name, class_idxs in processed_class_map.items():
            if class_name == "background":
                continue
            class_img = np.isin(img, class_idxs).astype(np.uint8)
            cc_img, num_labels = cc3d.connected_components(class_img, return_N=True)
            for instance_id in range(1, num_labels):
                _video_annotations = get_polygon_video_annotations(
                    cc_img,
                    class_idxs=[instance_id],
                    class_name=class_name,
                    slot_names=slot_names,
                    is_mpr=is_mpr,
                    pixdims_and_primary_planes=pixdims_and_primary_planes,
                    remote_file_path=remote_file_path,
                    legacy=legacy,
                )
                if _video_annotations:
                    video_annotations += _video_annotations
    elif mode == "video":  # For each class produce a single video annotation
        for class_name, class_idxs in processed_class_map.items():
            if class_name == "background":
                continue
            _video_annotations = get_polygon_video_annotations(
                img,
                class_idxs=class_idxs,
                class_name=class_name,
                slot_names=slot_names,
                is_mpr=is_mpr,
                pixdims_and_primary_planes=pixdims_and_primary_planes,
                remote_file_path=remote_file_path,
                legacy=legacy,
            )
            if _video_annotations is None:
                continue
            video_annotations += _video_annotations
    elif mode == "mask":
        video_annotations = get_mask_video_annotations(
            img,
            processed_class_map,
            slot_names,
            pixdims_and_primary_planes=pixdims_and_primary_planes,
            remote_file_path=remote_file_path,
            isotropic=legacy,
        )
    if mode in ["video", "instances"]:
        annotation_classes = {
            dt.AnnotationClass(class_name, "polygon", "polygon")
            for class_name in class_map.values()
        }
    elif mode == "mask":
        annotation_classes = {
            dt.AnnotationClass(class_name, "mask", "mask")
            for class_name in class_map.values()
        }
    remote_path = "/" if filename.parent == "." else filename.parent
    filename = Path(filename.name)
    return dt.AnnotationFile(
        path=json_path,
        filename=str(filename),
        remote_path=str(remote_path),
        annotation_classes=annotation_classes,
        annotations=video_annotations,
        slots=[
            dt.Slot(
                name=slot_name,
                type="dicom",
                source_files=[dt.SourceFile(file_name=str(filename), url=None)],
            )
            for slot_name in slot_names
        ],
    )


def get_polygon_video_annotations(
    volume: np.ndarray,
    class_name: str,
    class_idxs: List[int],
    slot_names: List[str],
    is_mpr: bool,
    pixdims_and_primary_planes: Dict[Path, Dict[str, Tuple[List[float], str]]],
    remote_file_path: Path,
    legacy: bool = False,
) -> Optional[List[dt.VideoAnnotation]]:
    if not is_mpr:
        if remote_file_path in pixdims_and_primary_planes:
            item_pixdims_and_primary_planes = pixdims_and_primary_planes[
                remote_file_path
            ]
            if slot_names:
                slot_name = slot_names[0]
            else:
                slot_name = list(item_pixdims_and_primary_planes.keys())[0]
            slot_pixims_and_primary_plane = item_pixdims_and_primary_planes[slot_name]
            primary_plane = slot_pixims_and_primary_plane[1]
            pixdims = slot_pixims_and_primary_plane[0]
        else:
            pixdims = [1.0, 1.0, 1.0]
            primary_plane = "AXIAL"

        if primary_plane == "AXIAL":
            view_idx = 2
        elif primary_plane == "CORONAL":
            view_idx = 1
        elif primary_plane == "SAGITTAL":
            view_idx = 0
        return nifti_to_video_polygon_annotation(
            volume,
            class_name,
            class_idxs,
            slot_names,
            pixdims,
            view_idx,
            legacy=legacy,
        )
    elif is_mpr and len(slot_names) == 3:
        video_annotations = []
        for view_idx, slot_name in enumerate(slot_names):
            pixdims = pixdims_and_primary_planes[remote_file_path][slot_name][0]
            _video_annotations = nifti_to_video_polygon_annotation(
                volume,
                class_name,
                class_idxs,
                [slot_name],
                pixdims,
                legacy=legacy,
                view_idx=view_idx,
            )
            video_annotations += _video_annotations
        return video_annotations
    else:
        raise Exception("If is_mpr is True, slot_names must be of length 3")


def get_mask_video_annotations(
    volume: np.ndarray,
    processed_class_map: Dict,
    slot_names: List[str],
    pixdims_and_primary_planes: Dict[Path, Dict[str, Tuple[List[float], str]]],
    remote_file_path: Path,
    isotropic: bool = False,
) -> Optional[List[dt.VideoAnnotation]]:
    """
    The function takes a volume and a class map and returns a list of video annotations

    We write a single raster layer for the volume but K mask annotations, where K is the number of classes.
    """
    if remote_file_path in pixdims_and_primary_planes:
        item_pixdims_and_primary_planess = pixdims_and_primary_planes[remote_file_path]
        if slot_names:
            slot_name = slot_names[0]
        else:
            slot_name = list(item_pixdims_and_primary_planess.keys())[0]
        slot_pixdims_and_primary_planes = item_pixdims_and_primary_planess[slot_name]
        primary_plane = slot_pixdims_and_primary_planes[1]
        pixdims = slot_pixdims_and_primary_planes[0]
    else:
        pixdims = [1.0, 1.0, 1.0]
        primary_plane = "AXIAL"

    frame_annotations = OrderedDict()
    all_mask_annotations = defaultdict(lambda: OrderedDict())
    # This is a dictionary of class_names to generated mask_annotation_ids
    mask_annotation_ids = {
        class_name: str(uuid.uuid4()) for class_name in processed_class_map.keys()
    }
    # We need to create a new mapping dictionary where the keys are the mask_annotation_ids
    # and the values are the new integers which we use in the raster layer

    if primary_plane == "AXIAL":
        view_idx = 2
    elif primary_plane == "CORONAL":
        view_idx = 1
    elif primary_plane == "SAGITTAL":
        view_idx = 0

    mask_annotation_ids_mapping = {}
    map_from_nifti_idx_to_raster_idx = {0: 0}  # 0 is the background class
    for i in range(volume.shape[view_idx]):
        if view_idx == 2:
            slice_mask = volume[:, :, i].astype(np.uint8)
        elif view_idx == 1:
            slice_mask = volume[:, i, :].astype(np.uint8)
        elif view_idx == 0:
            slice_mask = volume[i, :, :].astype(np.uint8)
        for raster_idx, (class_name, class_idxs) in enumerate(
            processed_class_map.items()
        ):
            if class_name == "background":
                continue
            class_mask = np.isin(slice_mask, class_idxs).astype(np.uint8).copy()
            if class_mask.sum() == 0:
                continue
            all_mask_annotations[class_name][i] = dt.make_mask(
                class_name, subs=None, slot_names=slot_names
            )
            all_mask_annotations[class_name][i].id = mask_annotation_ids[class_name]
            mask_annotation_ids_mapping[mask_annotation_ids[class_name]] = (
                raster_idx + 1
            )
            for class_idx in class_idxs:
                map_from_nifti_idx_to_raster_idx[class_idx] = raster_idx + 1
    # Now that we've created all the mask annotations, we need to create the raster layer
    # We only map the mask_annotation_ids which appear in any given frame.
    axial_size_after_isotropic_scaling = get_new_axial_size(volume, pixdims)
    for i in range(volume.shape[view_idx]):
        if view_idx == 2:
            slice_mask = volume[:, :, i].astype(np.uint8)
        elif view_idx == 1:
            slice_mask = volume[:, i, :].astype(np.uint8)
        elif view_idx == 0:
            slice_mask = volume[i, :, :].astype(np.uint8)
        if isotropic:
            slice_mask = zoom(
                slice_mask,
                (
                    axial_size_after_isotropic_scaling[0] / volume.shape[0],
                    axial_size_after_isotropic_scaling[1] / volume.shape[1],
                ),
            )

        # We need to convert from nifti_idx to raster_idx
        slice_mask = np.vectorize(
            lambda key: map_from_nifti_idx_to_raster_idx.get(key, 0)
        )(slice_mask)
        dense_rle = convert_to_dense_rle(slice_mask)
        raster_annotation = dt.make_raster_layer(
            class_name="__raster_layer__",
            mask_annotation_ids_mapping=mask_annotation_ids_mapping,
            total_pixels=slice_mask.size,
            dense_rle=dense_rle,
            slot_names=slot_names,
        )
        frame_annotations[i] = raster_annotation
    all_frame_ids = list(frame_annotations.keys())
    if not all_frame_ids:
        return None
    if len(all_frame_ids) == 1:
        segments = [[all_frame_ids[0], all_frame_ids[0] + 1]]
    elif len(all_frame_ids) > 1:
        segments = [[min(all_frame_ids), max(all_frame_ids)]]
    raster_video_annotation = dt.make_video_annotation(
        frame_annotations,
        keyframes={f_id: True for f_id in all_frame_ids},
        segments=segments,
        interpolated=False,
        slot_names=slot_names,
    )
    mask_video_annotations = []
    for class_name, mask_annotations in all_mask_annotations.items():
        mask_video_annotation = dt.make_video_annotation(
            mask_annotations,
            keyframes={f_id: True for f_id in all_frame_ids},
            segments=segments,
            interpolated=False,
            slot_names=slot_names,
        )
        mask_video_annotation.id = mask_annotation_ids[class_name]
        mask_video_annotations.append(mask_video_annotation)

    return [raster_video_annotation] + mask_video_annotations


def nifti_to_video_polygon_annotation(
    volume: np.ndarray,
    class_name: str,
    class_idxs: List[int],
    slot_names: List[str],
    pixdims: List[float],
    view_idx: int,
    legacy: bool = False,
) -> Optional[List[dt.VideoAnnotation]]:
    frame_annotations = OrderedDict()
    for i in range(volume.shape[view_idx]):
        if view_idx == 2:
            slice_mask = volume[:, :, i].astype(np.uint8)
            _pixdims = [pixdims[0], pixdims[1]]
        elif view_idx == 1:
            slice_mask = volume[:, i, :].astype(np.uint8)
            _pixdims = [pixdims[0], pixdims[2]]
        elif view_idx == 0:
            slice_mask = volume[i, :, :].astype(np.uint8)
            _pixdims = [pixdims[1], pixdims[2]]
        class_mask = np.isin(slice_mask, class_idxs).astype(np.uint8).copy()
        if class_mask.sum() == 0:
            continue
        polygon = mask_to_polygon(
            mask=class_mask,
            class_name=class_name,
            pixdims=_pixdims,
            legacy=legacy,
        )
        if polygon is None:
            continue
        frame_annotations[i] = polygon
    all_frame_ids = list(frame_annotations.keys())
    if not all_frame_ids:
        return None
    if len(all_frame_ids) == 1:
        segments = [[all_frame_ids[0], all_frame_ids[0] + 1]]
    elif len(all_frame_ids) > 1:
        segments = [[min(all_frame_ids), max(all_frame_ids) + 1]]
    video_annotation = dt.make_video_annotation(
        frame_annotations,
        keyframes={f_id: True for f_id in all_frame_ids},
        segments=segments,
        interpolated=False,
        slot_names=slot_names,
    )
    return [video_annotation]


def mask_to_polygon(
    mask: np.ndarray, class_name: str, pixdims: List[float], legacy: bool = False
) -> Optional[dt.Annotation]:
    def adjust_for_pixdims(x, y, pixdims):
        if legacy:
            if pixdims[1] > pixdims[0]:
                return {"x": y, "y": x * pixdims[1] / pixdims[0]}
            elif pixdims[1] < pixdims[0]:
                return {"x": y * pixdims[0] / pixdims[1], "y": x}
            else:
                return {"x": y, "y": x}
        else:
            return {"x": y * pixdims[0], "y": x * pixdims[1]}

    _labels, external_paths, _internal_paths = find_contours(mask)
    if len(external_paths) > 1:
        paths = []
        for external_path in external_paths:
            # skip paths with less than 2 points
            if len(external_path) // 2 <= 2:
                continue
            path = [
                adjust_for_pixdims(x, y, pixdims)
                for x, y in zip(external_path[0::2], external_path[1::2])
            ]
            paths.append(path)
        if len(paths) > 1:
            polygon = dt.make_polygon(class_name, paths)
        elif len(paths) == 1:
            polygon = dt.make_polygon(
                class_name,
                point_paths=paths[0],
            )
        else:
            return None
    elif len(external_paths) == 1:
        external_path = external_paths[0]
        if len(external_path) < 6:
            return None
        polygon = dt.make_polygon(
            class_name,
            point_paths=[
                adjust_for_pixdims(x, y, pixdims)
                for x, y in zip(external_path[0::2], external_path[1::2])
            ],
        )
    else:
        return None
    return polygon


def process_class_map(class_map):
    """
    This function takes a class_map and returns a dictionary with the class names as keys and
    all the corresponding class indexes as values.
    """
    processed_class_map = defaultdict(list)
    for key, value in class_map.items():
        processed_class_map[value].append(int(key))
    return processed_class_map


def rectify_header_sform_qform(img_nii):
    """
    Look at the sform and qform of the nifti object and correct it if any
    incompatibilities with pixel dimensions

    Adapted from https://github.com/NifTK/NiftyNet/blob/v0.6.0/niftynet/io/misc_io.py

    Args:
        img_nii: nifti image object
    """
    d = img_nii.header["dim"][0]
    pixdim = np.asarray(img_nii.header.get_zooms())[:d]
    sform, qform = img_nii.get_sform(), img_nii.get_qform()
    norm_sform = affine_to_spacing(sform, r=d)
    norm_qform = affine_to_spacing(qform, r=d)
    sform_mismatch = not np.allclose(norm_sform, pixdim)
    qform_mismatch = not np.allclose(norm_qform, pixdim)

    if img_nii.header["sform_code"] != 0:
        if not sform_mismatch:
            return img_nii
        if not qform_mismatch:
            img_nii.set_sform(img_nii.get_qform())
            return img_nii
    if img_nii.header["qform_code"] != 0:
        if not qform_mismatch:
            return img_nii
        if not sform_mismatch:
            img_nii.set_qform(img_nii.get_sform())
            return img_nii

    norm = affine_to_spacing(img_nii.affine, r=d)
    warnings.warn(f"Modifying image pixdim from {pixdim} to {norm}")

    img_nii.header.set_zooms(norm)
    return img_nii


def affine_to_spacing(
    affine: np.ndarray, r: int = 3, dtype=float, suppress_zeros: bool = True
) -> np.ndarray:
    """
    Copied over from monai.data.utils - https://docs.monai.io/en/stable/_modules/monai/data/utils.html

    Computing the current spacing from the affine matrix.

    Args:
        affine: a d x d affine matrix.
        r: indexing based on the spatial rank, spacing is computed from `affine[:r, :r]`.
        dtype: data type of the output.
        suppress_zeros: whether to surpress the zeros with ones.

    Returns:
        an `r` dimensional vector of spacing.
    """
    spacing = np.sqrt(np.sum(affine[:r, :r] * affine[:r, :r], axis=0))
    if suppress_zeros:
        spacing[spacing == 0] = 1.0
    return spacing


def correct_nifti_header_if_necessary(img_nii):
    """
    Check nifti object header's format, update the header if needed.
    In the updated image pixdim matches the affine.

    Args:
        img_nii: nifti image object
    """
    if img_nii.header.get("dim") is None:
        return img_nii  # not nifti?
    dim = img_nii.header["dim"][0]
    if dim >= 5:
        return img_nii  # do nothing for high-dimensional array
    # check that affine matches zooms
    pixdim = np.asarray(img_nii.header.get_zooms())[:dim]
    norm_affine = affine_to_spacing(img_nii.affine, r=dim)
    if np.allclose(pixdim, norm_affine):
        return img_nii
    if hasattr(img_nii, "get_sform"):
        return rectify_header_sform_qform(img_nii)
    return img_nii


def process_nifti(
    input_data: nib.nifti1.Nifti1Image,
    ornt: Optional[List[List[float]]] = [[0.0, -1.0], [1.0, -1.0], [2.0, -1.0]],
    remote_file_path: Path = Path("/"),
    legacy_remote_file_slot_affine_maps: Dict[Path, Dict[str, Any]] = {},
) -> Tuple[np.ndarray, Tuple[float], str]:
    """
    Converts a NifTI object of any orientation to the passed ornt orientation.
    The default ornt is LPI.

    Files that were uploaded before the `MED_2D_VIEWER` feature are `legacy`. Non-legacy
    files are uploaded and re-oriented to the `LPI` orientation. Legacy files
    files were treated differently:
    - Legacy NifTI files were re-oriented to `LPI`, but their
      affine was stored as `RAS`, which is the opposite orientation. However, because
      their pixel data is stored in `LPI`, we can treat them the same way as non-legacy
      files.
    - Legacy DICOM files were not always re-oriented to `LPI`. We therefore use the
      affine of the dataset item from `slot_affine_map` to re-orient the NifTI file to
      be imported

    Args:
        input_data: nibabel NifTI object.
        ornt: (n,2) orientation array. It defines a transformation to LPI
            ornt[N,1] is a flip of axis N of the array, where 1 means no flip and -1 means flip.
            ornt[:,0] is the transpose that needs to be done to the implied array, as in arr.transpose(ornt[:,0]).
        remote_file_path: Path
            The full path of the remote file
        legacy_remote_file_slot_affine_maps: Dict[Path, Dict[str, Any]]
            A dictionary of remote file full paths to their slot affine maps

    Returns:
        data_array: pixel array with orientation ornt.
    """
    img = correct_nifti_header_if_necessary(input_data)
    orig_ax_codes = nib.orientations.aff2axcodes(img.affine)
    orig_ornt = nib.orientations.axcodes2ornt(orig_ax_codes)
    is_dicom = remote_file_path.suffix.lower() == ".dcm"
    if remote_file_path in legacy_remote_file_slot_affine_maps and is_dicom:
        slot_affine_map = legacy_remote_file_slot_affine_maps[remote_file_path]
        affine = slot_affine_map[next(iter(slot_affine_map))]  # Take the 1st slot
        ax_codes = nib.orientations.aff2axcodes(affine)
        ornt = nib.orientations.axcodes2ornt(ax_codes)
    transform = nib.orientations.ornt_transform(orig_ornt, ornt)
    reoriented_img = img.as_reoriented(transform)
    data_array = reoriented_img.get_fdata()
    return data_array


def convert_to_dense_rle(raster: np.ndarray) -> List[int]:
    dense_rle, prev_val, cnt = [], None, 0
    for val in raster.T.flat:
        if val == prev_val:
            cnt += 1
        else:
            if prev_val is not None:
                dense_rle.extend([int(prev_val), int(cnt)])
            prev_val, cnt = val, 1
    dense_rle.extend([int(prev_val), int(cnt)])
    return dense_rle


def get_new_axial_size(
    volume: np.ndarray,
    pixdims: List[float],
    isotropic: bool = False,
) -> Tuple[int, int]:
    """Get the new size of the Axial plane after resizing to isotropic pixel dimensions.

    Args:
        volume: Input volume.
        pixdims: The pixel dimensions / spacings of the volume.
        isotropic: bool, default: True
            If True, the function will attempt to resize the annotations to isotropic pixel dimensions.
            If False, the function will not attempt to resize the annotations to isotropic pixel dimensions.

    Returns:
        Tuple[int, int]: The new size of the Axial plane.
    """
    original_size = volume.shape
    if not isotropic:
        return original_size[0], original_size[1]

    original_spacing = pixdims
    min_spacing = min(pixdims[0], pixdims[1])
    return (
        int(round(original_size[0] * (original_spacing[0] / min_spacing))),
        int(round(original_size[1] * (original_spacing[1] / min_spacing))),
    )
